from typing import List

import numpy as np
import optuna
import panel as pn
import param
from optuna.visualization import (
    plot_param_importances,
    plot_pareto_front,
    plot_optimization_history,
)

from bencher.bench_cfg import BenchCfg


from bencher.variables.inputs import IntSweep, FloatSweep, StringSweep, EnumSweep, BoolSweep
from bencher.variables.time import TimeSnapshot, TimeEvent
from bencher.variables.results import OptDir

from bencher.variables.parametrised_sweep import ParametrizedSweep


def optuna_grid_search(bench_cfg: BenchCfg) -> optuna.Study:
    """use optuna to perform a grid search

    Args:
        bench_cfg (BenchCfg): setting for grid search

    Returns:
        optuna.Study: results of grid search
    """
    search_space = {}
    for iv in bench_cfg.all_vars:
        search_space[iv.name] = iv.values(bench_cfg.debug)
    directions = []
    for rv in bench_cfg.result_vars:
        if rv.direction != OptDir.none:
            directions.append(rv.direction)

    study = optuna.create_study(
        sampler=optuna.samplers.GridSampler(search_space),
        directions=directions,
        study_name=bench_cfg.title,
    )
    return study


# def bench_results_to_optuna(bench_cfg: BenchCfg):


def to_optuna(worker, bench_cfg: BenchCfg, n_trials=100, sampler=optuna.samplers.TPESampler()):
    directions = []
    for rv in bench_cfg.result_vars:
        if rv.direction != OptDir.none:
            directions.append(rv.direction)

    study = optuna.create_study(sampler=sampler, directions=directions, study_name=bench_cfg.title)

    # add already calculated results
    if len(bench_cfg.ds.sizes) > 0:
        study.add_trials(bench_results_to_optuna_trials(bench_cfg, True))

    def wrapped(trial):
        kwargs = {}
        for iv in bench_cfg.input_vars:
            kwargs[iv.name] = sweep_var_to_suggest(iv, trial)
        result = worker(**kwargs)
        output = []
        for rv in bench_cfg.result_vars:
            output.append(result[rv.name])
        return tuple(output)

    study.optimize(wrapped, n_trials=n_trials)
    return study


def extract_study_to_dataset(study: optuna.Study, bench_cfg: BenchCfg) -> BenchCfg:
    """Extract an optuna study into an xarray dataset for easy plotting

    Args:
        study (optuna.Study): The result of a gridsearch
        bench_cfg (BenchCfg): Options for the grid search

    Returns:
        BenchCfg: An updated config with the results included
    """
    for t in study.trials:
        for it, rv in enumerate(bench_cfg.result_vars):
            bench_cfg.ds[rv.name].loc[t.params] = t.values[it]
    return bench_cfg


def param_importance(bench_cfg: BenchCfg, study: optuna.Study) -> pn.Row:
    col_importance = pn.Column()
    for tgt in bench_cfg.optuna_targets():
        col_importance.append(
            pn.Column(
                pn.pane.Markdown(f"## Parameter importance for: {tgt}"),
                plot_param_importances(study, target=lambda t: t.values[0], target_name=tgt),
            )
        )
    return col_importance


def collect_optuna_plots(bench_cfg: BenchCfg) -> List[pn.pane.panel]:
    """Use optuna to plot various summaries of the optimisation

    Args:
        study (optuna.Study): The study to plot
        bench_cfg (BenchCfg): Benchmark config with options used to generate the study

    Returns:
        List[pn.pane.Pane]: A list of plots
    """

    studies = [bench_cfg_to_study(bench_cfg, True)]
    bench_cfg.studies = studies
    titles = ["# Analysis"]
    if bench_cfg.repeats > 1:
        studies.append(bench_cfg_to_study(bench_cfg, False))
        titles = ["# Parameter Importance With Repeats", "# Parameter Importance Without Repeats"]

    study_repeats_pane = pn.Row()
    for study, title in zip(studies, titles):
        study_pane = pn.Column()
        target_names = bench_cfg.optuna_targets()
        param_str = []

        study_pane.append(pn.pane.Markdown(title))

        if len(target_names) > 1:
            if len(target_names) <= 3:
                study_pane.append(
                    plot_pareto_front(
                        study, target_names=target_names, include_dominated_trials=False
                    )
                )
            else:
                print("plotting pareto front of first 3 result variables")
                study_pane.append(
                    plot_pareto_front(
                        study,
                        targets=lambda t: (t.values[0], t.values[1], t.values[2]),
                        target_names=target_names[:3],
                        include_dominated_trials=False,
                    )
                )

            study_pane.append(param_importance(bench_cfg, study))
            param_str.append(f"    Number of trials on the Pareto front: {len(study.best_trials)}")
            for t in study.best_trials:
                param_str.extend(summarise_trial(t, bench_cfg))

        else:
            # cols.append(plot_optimization_history(study)) #TODO, maybe more clever when this is plotted?

            # If there is only 1 parameter then there is no point is plotting relative importance.  Only worth plotting if there are multiple repeats of the same value so that you can compare the parameter vs to repeat to get a sense of the how much chance affects the results
            # if bench_cfg.repeats > 1 and len(bench_cfg.input_vars) > 1:  #old code, not sure if its right
            if len(bench_cfg.input_vars) > 1:
                study_pane.append(plot_param_importances(study, target_name=target_names[0]))

            param_str.extend(summarise_trial(study.best_trial, bench_cfg))

        kwargs = {"height": 500, "scroll": True} if len(param_str) > 30 else {}

        param_str = "\n".join(param_str)
        study_pane.append(
            pn.Row(pn.pane.Markdown(f"## Best Parameters\n```text\n{param_str}"), **kwargs),
        )

        study_repeats_pane.append(study_pane)

    return study_repeats_pane


def summarise_trial(trial: optuna.trial, bench_cfg: BenchCfg) -> List[str]:
    """Given a trial produce a string summary of the best results

    Args:
        trial (optuna.trial): trial to summarise
        bench_cfg (BenchCfg): info about the trial

    Returns:
        List[str]: Summary of trial
    """
    sep = "    "
    output = []
    output.append(f"Trial id:{trial.number}:")
    output.append(f"{sep}Inputs:")
    for k, v in trial.params.items():
        output.append(f"{sep}{sep}{k}:{v}")
    output.append(f"{sep}Results:")
    for it, rv in enumerate(bench_cfg.result_vars):
        if rv.direction != OptDir.none:
            output.append(f"{sep}{sep}{rv.name}:{trial.values[it]}")
    return output


def sweep_var_to_optuna_dist(var: param.Parameter) -> optuna.distributions.BaseDistribution:
    """Convert a sweep var to an optuna distribution

    Args:
        var (param.Parameter): A sweep var

    Raises:
        ValueError: Unsupported var type

    Returns:
        optuna.distributions.BaseDistribution: Optuna representation of a sweep var
    """

    iv_type = type(var)
    if iv_type == IntSweep:
        return optuna.distributions.IntDistribution(var.bounds[0], var.bounds[1])
    if iv_type == FloatSweep:
        return optuna.distributions.FloatDistribution(var.bounds[0], var.bounds[1])
    if iv_type in (EnumSweep, StringSweep):
        return optuna.distributions.CategoricalDistribution(var.objects)
    if iv_type == BoolSweep:
        return optuna.distributions.CategoricalDistribution([False, True])
    if iv_type == TimeSnapshot:
        # return optuna.distributions.IntDistribution(0, sys.maxsize)
        return optuna.distributions.FloatDistribution(0, 1e20)
        # return optuna.distributions.CategoricalDistribution([])
    # elif iv_type == TimeEvent:
    #     pass
    # return optuna.distributions.CategoricalDistribution(["now"])

    raise ValueError(f"This input type {iv_type} is not supported")


def sweep_var_to_suggest(iv: ParametrizedSweep, trial: optuna.trial) -> object:
    """Converts from a sweep var to an optuna

    Args:
        iv (ParametrizedSweep): A parametrized sweep input variable
        trial (optuna.trial): Optuna trial used to define the sample

    Raises:
        ValueError: Unsupported var type

    Returns:
        Any: A sampled variable (can be any type)
    """
    iv_type = type(iv)

    if iv_type == IntSweep:
        return trial.suggest_int(iv.name, iv.bounds[0], iv.bounds[1])
    if iv_type == FloatSweep:
        return trial.suggest_float(iv.name, iv.bounds[0], iv.bounds[1])
    if iv_type in (EnumSweep, StringSweep):
        return trial.suggest_categorical(iv.name, iv.objects)
    if iv_type in (TimeSnapshot, TimeEvent):
        pass  # optuna does not like time
    if iv_type == BoolSweep:
        return trial.suggest_categorical(iv.name, [True, False])
    raise ValueError(f"This input type {iv_type} is not supported")


def cfg_from_optuna_trial(
    trial: optuna.trial, bench_cfg: BenchCfg, cfg_type: ParametrizedSweep
) -> ParametrizedSweep:
    cfg = cfg_type()
    for iv in bench_cfg.input_vars:
        cfg.param.set_param(iv.name, sweep_var_to_suggest(iv, trial))
    for mv in bench_cfg.meta_vars:
        sweep_var_to_suggest(mv, trial)
    return cfg


def bench_results_to_optuna_trials(bench_cfg: BenchCfg, include_meta: bool = True) -> optuna.Study:
    """Convert an xarray dataset to an optuna study so optuna can further optimise or plot the statespace

    Args:
        bench_cfg (BenchCfg): benchmark config to convert

    Returns:
        optuna.Study: optuna description of the study
    """
    if include_meta:
        df = bench_cfg.get_dataframe()
        all_vars = []
        for v in bench_cfg.all_vars:
            if type(v) != TimeEvent:
                all_vars.append(v)

        print("All vars", all_vars)
    else:
        all_vars = bench_cfg.input_vars
        df = bench_cfg.ds.mean("repeat").to_dataframe().reset_index()
    # df = bench_cfg.ds.mean("repeat").to_dataframe.reset_index()
    # bench_cfg.all_vars
    # del bench_cfg.meta_vars[1]

    trials = []
    distributions = {}
    for i in all_vars:
        distributions[i.name] = sweep_var_to_optuna_dist(i)

    for row in df.iterrows():
        params = {}
        values = []
        for i in all_vars:
            if type(i) == TimeSnapshot:
                if type(row[1][i.name]) == np.datetime64:
                    params[i.name] = row[1][i.name].timestamp()
            else:
                params[i.name] = row[1][i.name]

        for r in bench_cfg.result_vars:
            if r.direction != OptDir.none:
                values.append(row[1][r.name])

        trials.append(
            optuna.trial.create_trial(
                params=params,
                distributions=distributions,
                values=values,
            )
        )
    return trials


def bench_cfg_to_study(bench_cfg: BenchCfg, include_meta: bool) -> optuna.Study:
    trials = bench_results_to_optuna_trials(bench_cfg, include_meta)
    study = optuna_grid_search(bench_cfg)
    optuna.logging.set_verbosity(optuna.logging.CRITICAL)
    import warnings

    # /usr/local/lib/python3.10/dist-packages/optuna/samplers/_grid.py:224: UserWarning: over_time contains a value with the type of <class 'pandas._libs.tslibs.timestamps.Timestamp'>, which is not supported by `GridSampler`. Please make sure a value is `str`, `int`, `float`, `bool` or `None` for persistent storage.

    # this is not disabling the warning
    warnings.filterwarnings(action="ignore", category=UserWarning)
    # remove optuna gridsearch warning as we are not using their gridsearch because it has the most inexplicably terrible performance I have ever seen in my life. How can a for loop of 400 iterations start out with 100ms per loop and increase to greater than a 1000ms after 250ish iterations!?!?!??!!??!
    study.add_trials(trials)
    return study


def summarise_study(study: optuna.study.Study) -> None:
    row = pn.Column(name="Optimisation Results")
    row.append(plot_optimization_history(study))
    row.append(plot_param_importances(study))
    try:
        row.append(plot_pareto_front(study))
    except Exception:
        pass

    row.append(
        pn.pane.Markdown(f"```\nBest value: {study.best_value}\nParams: {study.best_params}```")
    )

    return row



import logging
from datetime import datetime
from itertools import product
from typing import Callable, List
from copy import deepcopy
import os
import numpy as np
import panel as pn
import param
import xarray as xr
from diskcache import Cache
from sortedcontainers import SortedDict
from contextlib import suppress

from bencher.bench_cfg import BenchCfg, BenchRunCfg, DimsCfg
from bencher.bench_plot_server import BenchPlotServer


from bencher.variables.sweep_base import hash_sha1
from bencher.variables.inputs import IntSweep
from bencher.variables.time import TimeSnapshot, TimeEvent
from bencher.variables.results import ResultVar, ResultVec

from bencher.variables.parametrised_sweep import ParametrizedSweep

from bencher.plotting.plot_collection import PlotCollection
from bencher.plt_cfg import BenchPlotter
from bencher.plotting.plot_library import PlotLibrary  # noqa pylint: disable=unused-import
from bencher.utils import hmap_canonical_input

from bencher.optuna_conversions import to_optuna, summarise_study
from optuna import Study
from pathlib import Path
import shutil


# Customize the formatter
formatter = logging.Formatter("%(levelname)s: %(message)s")
logging.basicConfig(level=logging.INFO, format="%(levelname)s %(message)s")

for handler in logging.root.handlers:
    handler.setFormatter(formatter)


def set_xarray_multidim(data_array: xr.DataArray, index_tuple, value: float) -> xr.DataArray:
    """This is terrible, I need to do this in a better way, but [] does not like *args syntax and the () version of the set function doesn't either"""
    match len(index_tuple):
        case 1:
            data_array[index_tuple[0]] = value
        case 2:
            data_array[index_tuple[0], index_tuple[1]] = value
        case 3:
            data_array[index_tuple[0], index_tuple[1], index_tuple[2]] = value
        case 4:
            data_array[index_tuple[0], index_tuple[1], index_tuple[2], index_tuple[3]] = value
        case 5:
            data_array[
                index_tuple[0], index_tuple[1], index_tuple[2], index_tuple[3], index_tuple[4]
            ] = value
        case 6:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
            ] = value
        case 7:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
                index_tuple[6],
            ] = value
        case 8:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
                index_tuple[6],
                index_tuple[7],
            ] = value
        case 9:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
                index_tuple[6],
                index_tuple[7],
                index_tuple[8],
            ] = value
    return data_array


class Bench(BenchPlotServer):
    def __init__(
        self,
        bench_name: str = None,
        worker: Callable | ParametrizedSweep = None,
        worker_input_cfg: ParametrizedSweep = None,
        plot_lib: PlotCollection = None,
        remove_plots: list = None,
        run_cfg=None,
    ) -> None:
        """Create a new Bench object from a function and a class defining the inputs to the function

        Args:
            bench_name (str): The name of the benchmark and output folder for the figures
            worker (Callable | ParametrizedSweep): A function that accepts a class of type (worker_input_config)
            worker_input_config (ParametrizedSweep): A class defining the parameters of the function.
            plot_lib: (PlotCollection):  A dictionary of plot names:method pairs that are selected for plotting based on the type of data they can plot.
        """
        self.bench_name = bench_name
        self.worker = None
        self.worker_class_instance = None
        self.worker_input_cfg = None
        self.worker_class_instance = None
        self.set_worker(worker, worker_input_cfg)
        self.run_cfg = run_cfg

        self.pane = pn.Tabs(tabs_location="left", name=self.bench_name)
        # The number of times the wrapped worker was called
        self.worker_wrapper_call_count = 0
        self.worker_fn_call_count = 0  # The number of times the raw worker was called
        # The number of times the cache was used instead of the raw worker
        self.worker_cache_call_count = 0
        self.bench_cfg_hashes = []  # a list of hashes that point to benchmark results
        self.last_run_cfg = None  # cached run_cfg used to pass to the plotting function
        self.sample_cache = None  # store the results of each benchmark function call in a cache
        self.ds_dynamic = {}  # A dictionary to store unstructured vector datasets
        self.plot_lib = PlotLibrary.default() if plot_lib is None else plot_lib
        self.cache_size = int(100e9)  # default to 100gb
        if remove_plots is not None:
            for i in remove_plots:
                self.plot_lib.remove(i)

    def set_worker(self, worker: Callable, worker_input_cfg: ParametrizedSweep = None) -> None:
        """Set the benchmark worker function and optionally the type the worker expects

        Args:
            worker (Callable): The benchmark worker function
            worker_input_cfg (ParametrizedSweep, optional): The input type the worker expects. Defaults to None.
        """
        if isinstance(worker, ParametrizedSweep):
            # if issubclass(worker,ParametrizedSweep):
            # logging.warning("This should be a class instance, not a class")
            self.worker_class_instance = worker
            self.worker = self.worker_class_instance.__call__
            logging.info("setting worker from bench class.__call__")
        else:
            self.worker = worker
            logging.info(f"setting worker {worker}")
        self.worker_input_cfg = worker_input_cfg

    def to_optuna(
        self,
        input_vars: List[ParametrizedSweep],
        result_vars: List[ParametrizedSweep],
        n_trials: int = 100,
    ) -> Study:
        bench_cfg = BenchCfg(
            input_vars=input_vars,
            result_vars=result_vars,
            bench_name=self.bench_name,
        )
        return self.to_optuna_from_sweep(bench_cfg, n_trials)

    def to_optuna_from_sweep(
        self,
        bench_cfg: BenchCfg,
        n_trials: int = 100,
    ) -> Study:
        optu = to_optuna(self.worker, bench_cfg, n_trials=n_trials)
        self.append(summarise_study(optu))
        return optu

    def plot_sweep(
        self,
        title: str,
        input_vars: List[ParametrizedSweep] = None,
        result_vars: List[ParametrizedSweep] = None,
        const_vars: List[ParametrizedSweep] = None,
        time_src: datetime = None,
        description: str = None,
        post_description: str = None,
        pass_repeat: bool = False,
        tag: str = "",
        run_cfg: BenchRunCfg = None,
        plot_lib=None,
    ) -> BenchCfg:
        """The all in 1 function benchmarker and results plotter.

        Args:
            input_vars (List[ParametrizedSweep], optional): _description_. Defaults to None.
            result_vars (List[ParametrizedSweep], optional): _description_. Defaults to None.
            const_vars (List[ParametrizedSweep], optional): A list of variables to keep constant with a specified value. Defaults to None.
            title (str, optional): The title of the benchmark. Defaults to None.
            description (str, optional): A description of the benchmark. Defaults to None.
            post_description (str, optional): A description that comes after the benchmark plots. Defaults to None.
            time_src (datetime, optional): Set a time that the result was generated. Defaults to datetime.now().
            pass_repeat (bool,optional) By default do not pass the kwarg 'repeat' to the benchmark function.  Set to true if
            you want the benchmark function to be passed the repeat number
            tag (str,optional): Use tags to group different benchmarks together.
            run_cfg: (BenchRunCfg, optional): A config for storing how the benchmarks and run and plotted
            plot_lib: (PlotCollection):  A dictionary of plot names:method pairs that are selected for plotting based on the type of data they can plot.

        Raises:
            ValueError: If a result variable is not set

        Returns:
            BenchCfg: A class with all the data used to generate the results and the results
        """

        if self.worker_class_instance is not None:
            if input_vars is None:
                logging.info(
                    "No input variables passed, using all param variables in bench class as inputs"
                )
                input_vars = self.worker_class_instance.get_inputs_only()
                for i in input_vars:
                    logging.info(i.name)
            if result_vars is None:
                logging.info(
                    "No results variables passed, using all result variables in bench class:"
                )
                result_vars = self.worker_class_instance.get_results_only()
                for r in result_vars:
                    logging.info(f"result var: {r.name}")
            if const_vars is None:
                const_vars = self.worker_class_instance.get_input_defaults()
        else:
            if input_vars is None:
                input_vars = []
            if result_vars is None:
                result_vars = []
            if const_vars is None:
                const_vars = []
            else:
                const_vars = deepcopy(const_vars)

        if run_cfg is None:
            if self.run_cfg is None:
                run_cfg = BenchRunCfg()
                logging.info("Generate default run cfg")
            else:
                run_cfg = deepcopy(self.run_cfg)
                logging.info("Copy run cfg from bench class")
        if run_cfg.only_plot:
            run_cfg.use_cache = True

        self.last_run_cfg = run_cfg

        if run_cfg.level > 0:
            inputs = []
            for i in input_vars:
                inputs.append(i.with_level(run_cfg.level))
            input_vars = inputs

        # if any of the inputs have been include as constants, remove those variables from the list of constants
        with suppress(ValueError, AttributeError):
            for i in input_vars:
                for c in const_vars:
                    # print(i.hash_persistent())
                    if i.name == c[0].name:
                        const_vars.remove(c)
                        logging.info(f"removing {i.name} from constants")

        for i in input_vars:
            self.check_var_is_a_param(i, "input")
        for i in result_vars:
            self.check_var_is_a_param(i, "result")
        for i in const_vars:
            # consts come as tuple pairs
            self.check_var_is_a_param(i[0], "const")

        if post_description is None:
            post_description = (
                "## Results Description\nPlease set post_description to explain these results"
            )

        bench_cfg = BenchCfg(
            input_vars=input_vars,
            result_vars=result_vars,
            const_vars=const_vars,
            bench_name=self.bench_name,
            description=description,
            post_description=post_description,
            title=title,
            pass_repeat=pass_repeat,
            tag=run_cfg.run_tag + tag,
        )
        print("tag", bench_cfg.tag)

        bench_cfg.param.update(run_cfg.param.values())
        bench_cfg.plot_lib = plot_lib if plot_lib is not None else self.plot_lib

        print("plot_lib", bench_cfg.plot_lib)

        bench_cfg_hash = bench_cfg.hash_persistent(True)
        bench_cfg.hash_value = bench_cfg_hash

        # does not include repeats in hash as sample_hash already includes repeat as part of the per sample hash
        bench_cfg_sample_hash = bench_cfg.hash_persistent(False)

        if bench_cfg.use_sample_cache:
            self.sample_cache = Cache(
                "cachedir/sample_cache", tag_index=True, size_limit=self.cache_size
            )
            if bench_cfg.clear_sample_cache:
                self.clear_tag_from_cache(bench_cfg.tag)

        calculate_results = True
        with Cache("cachedir/benchmark_inputs", size_limit=self.cache_size) as c:
            if run_cfg.clear_cache:
                c.delete(bench_cfg_hash)
                logging.info("cleared cache")
            elif run_cfg.use_cache:
                logging.info(
                    f"checking for previously calculated results with key: {bench_cfg_hash}"
                )
                if bench_cfg_hash in c:
                    logging.info(f"loading cached results from key: {bench_cfg_hash}")
                    bench_cfg = c[bench_cfg_hash]
                    # if not over_time:  # if over time we always want to calculate results
                    calculate_results = False
                else:
                    logging.info("did not detect results in cache")
                    if run_cfg.only_plot:
                        raise FileNotFoundError("Was not able to load the results to plot!")

        if calculate_results:
            if run_cfg.time_event is not None:
                time_src = run_cfg.time_event
            bench_cfg = self.calculate_benchmark_results(
                bench_cfg, time_src, bench_cfg_sample_hash, run_cfg
            )
            if self.sample_cache is not None:
                self.sample_cache.close()

            # use the hash of the inputs to look up historical values in the cache
            if run_cfg.over_time:
                bench_cfg.ds = self.load_history_cache(
                    bench_cfg.ds, bench_cfg_hash, run_cfg.clear_history
                )

            self.report_results(bench_cfg, run_cfg.print_xarray, run_cfg.print_pandas)
            self.cache_results(bench_cfg, bench_cfg_hash)

        if self.sample_cache is not None:
            logging.info(
                f"cache size :{int(self.sample_cache.volume() / 1000000)}MB / {int(self.cache_size/1000000)}MB"
            )

        self.pane = BenchPlotter.plot(bench_cfg, self.pane)
        return bench_cfg

    def check_var_is_a_param(self, variable: param.Parameter, var_type: str):
        """check that a variable is a subclass of param

        Args:
            variable (param.Parameter): the varible to check
            var_type (str): a string representation of the variable type for better error messages

        Raises:
            TypeError: the input variable type is not a param.
        """
        if not isinstance(variable, param.Parameter):
            raise TypeError(
                f"You need to use {var_type}_vars =[{self.worker_input_cfg}.param.your_variable], instead of {var_type}_vars =[{self.worker_input_cfg}.your_variable]"
            )

    def cache_results(self, bench_cfg: BenchCfg, bench_cfg_hash: int) -> None:
        with Cache("cachedir/benchmark_inputs", size_limit=self.cache_size) as c:
            logging.info(f"saving results with key: {bench_cfg_hash}")
            self.bench_cfg_hashes.append(bench_cfg_hash)
            c[bench_cfg_hash] = bench_cfg
            logging.info(f"saving benchmark: {self.bench_name}")
            c[self.bench_name] = self.bench_cfg_hashes

    def calculate_benchmark_results(
        self, bench_cfg, time_src: datetime | str, bench_cfg_sample_hash, bench_run_cfg
    ):
        """A function for generating an n-d xarray from a set of input variables in the BenchCfg

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            time_src (datetime): a representation of the sample time

        Returns:
            bench_cfg (BenchCfg): description of the benchmark parameters
        """
        bench_cfg, func_inputs, dims_name = self.setup_dataset(bench_cfg, time_src)
        constant_inputs = self.define_const_inputs(bench_cfg.const_vars)
        callcount = 1
        bench_cfg.hmap_kdims = sorted(dims_name)

        for idx_tuple, function_input_vars in func_inputs:
            logging.info(f"{bench_cfg.title}:call {callcount}/{len(func_inputs)}")
            self.call_worker_and_store_results(
                bench_cfg,
                idx_tuple,
                function_input_vars,
                dims_name,
                constant_inputs,
                bench_cfg_sample_hash,
                bench_run_cfg,
            )
            callcount += 1

        for inp in bench_cfg.all_vars:
            self.add_metadata_to_dataset(bench_cfg, inp)
        return bench_cfg

    def show(self, run_cfg: BenchRunCfg = None) -> None:
        """Launches a webserver with plots of the benchmark results, blocking

        Args:
            run_cfg (BenchRunCfg, optional): Options for the webserve such as the port. Defaults to None.

        """
        if run_cfg is None:
            if self.last_run_cfg is not None:
                run_cfg = self.last_run_cfg
            else:
                run_cfg = BenchRunCfg()

        BenchPlotServer().plot_server(self.bench_name, run_cfg, self.pane)

    def load_history_cache(
        self, ds: xr.Dataset, bench_cfg_hash: int, clear_history: bool
    ) -> xr.Dataset:
        """Load historical data from a cache if over_time=true

        Args:
            ds (xr.Dataset): Freshly calcuated data
            bench_cfg_hash (int): Hash of the input variables used to generate the data
            clear_history (bool): Optionally clear the history

        Returns:
            xr.Dataset: historical data as an xr dataset
        """
        with Cache("cachedir/history", size_limit=self.cache_size) as c:
            if clear_history:
                logging.info("clearing history")
            else:
                logging.info(f"checking historical key: {bench_cfg_hash}")
                if bench_cfg_hash in c:
                    logging.info("loading historical data from cache")
                    ds_old = c[bench_cfg_hash]
                    ds = xr.concat([ds_old, ds], "over_time")
                else:
                    logging.info("did not detect any historical data")

            logging.info("saving data to history cache")
            c[bench_cfg_hash] = ds
        return ds

    def setup_dataset(
        self, bench_cfg: BenchCfg, time_src: datetime | str
    ) -> tuple[BenchCfg, List, List]:
        """A function for generating an n-d xarray from a set of input variables in the BenchCfg

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            time_src (datetime | str): a representation of the sample time

        Returns:
            _type_: _description_
        """

        if time_src is None:
            time_src = datetime.now()
        bench_cfg.meta_vars = self.define_extra_vars(bench_cfg, bench_cfg.repeats, time_src)

        bench_cfg.all_vars = bench_cfg.input_vars + bench_cfg.meta_vars

        for i in bench_cfg.all_vars:
            logging.info(i.sampling_str(bench_cfg.debug))

        dims_cfg = DimsCfg(bench_cfg)
        function_inputs = list(
            zip(product(*dims_cfg.dim_ranges_index), product(*dims_cfg.dim_ranges))
        )
        # xarray stores K N-dimensional arrays of data.  Each array is named and in this case we have a nd array for each result variable
        data_vars = {}

        for rv in bench_cfg.result_vars:
            if type(rv) == ResultVar:
                result_data = np.empty(dims_cfg.dims_size)
                result_data.fill(np.nan)
                data_vars[rv.name] = (dims_cfg.dims_name, result_data)
            elif type(rv) == ResultVec:
                for i in range(rv.size):
                    result_data = np.full(dims_cfg.dims_size, np.nan)
                    data_vars[rv.index_name(i)] = (dims_cfg.dims_name, result_data)

        bench_cfg.ds = xr.Dataset(data_vars=data_vars, coords=dims_cfg.coords)
        bench_cfg.ds_dynamic = self.ds_dynamic

        return bench_cfg, function_inputs, dims_cfg.dims_name

    def define_const_inputs(self, const_vars) -> dict:
        constant_inputs = None
        if const_vars is not None:
            const_vars, constant_values = [
                [i for i, j in const_vars],
                [j for i, j in const_vars],
            ]

            constant_names = [i.name for i in const_vars]
            constant_inputs = dict(zip(constant_names, constant_values))
        return constant_inputs

    def define_extra_vars(self, bench_cfg: BenchCfg, repeats: int, time_src) -> list[IntSweep]:
        """Define extra meta vars that are stored in the n-d array but are not passed to the benchmarking function, such as number of repeats and the time the function was called.

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            repeats (int): the number of times to sample the function
            time_src (datetime): a representation of the sample time

        Returns:
            _type_: _description_
        """
        bench_cfg.iv_repeat = IntSweep(
            default=repeats,
            bounds=[1, repeats],
            samples=repeats,
            samples_debug=2 if repeats > 2 else 1,
            units="repeats",
        )
        bench_cfg.iv_repeat.name = "repeat"
        extra_vars = [bench_cfg.iv_repeat]

        if bench_cfg.over_time:
            if isinstance(time_src, str):
                iv_over_time = TimeEvent(time_src)
            else:
                iv_over_time = TimeSnapshot(time_src)
            iv_over_time.name = "over_time"
            extra_vars.append(iv_over_time)
            bench_cfg.iv_time = [iv_over_time]
        return extra_vars

    def worker_wrapper(self, bench_cfg: BenchCfg, function_input: dict):
        self.worker_fn_call_count += 1
        if not bench_cfg.pass_repeat:
            function_input.pop("repeat")
        if "over_time" in function_input:
            function_input.pop("over_time")
        if "time_event" in function_input:
            function_input.pop("time_event")

        if self.worker_input_cfg is None:  # worker takes kwargs
            return self.worker(**function_input)

        # worker takes a parametrised input object
        input_cfg = self.worker_input_cfg()
        for k, v in function_input.items():
            input_cfg.param.set_param(k, v)

        return self.worker(input_cfg)

    def call_worker_and_store_results(
        self,
        bench_cfg: BenchCfg,
        index_tuple: tuple,
        function_input_vars: List,
        dims_name: List[str],
        constant_inputs: dict,
        bench_cfg_sample_hash: str,
        bench_run_cfg: BenchRunCfg,
    ) -> None:
        """A wrapper around the benchmarking function to set up and store the results of the benchmark function

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            index_tuple (tuple): tuple of n-d array indices
            function_input_vars (List): list of function inputs as dicts
            dims_name (List[str]): names of the n-d dimension
            constant_inputs (dict): input values to keep constant
        """
        self.worker_wrapper_call_count += 1
        function_input = SortedDict(zip(dims_name, function_input_vars))

        canonical_input = hmap_canonical_input(function_input)
        if constant_inputs is not None:
            function_input = function_input | constant_inputs

        if bench_cfg.print_bench_inputs:
            logging.info("Bench Inputs:")
            for k, v in function_input.items():
                logging.info(f"\t {k}:{v}")

        # store a tuple of the inputs as keys for a holomap
        if bench_cfg.use_sample_cache:
            # the signature is the hash of the inputs to to the function + meta variables such as repeat and time + the hash of the benchmark sweep as a whole (without the repeats hash)
            fn_inputs_sorted = list(SortedDict(function_input).items())
            function_input_signature_pure = hash_sha1((fn_inputs_sorted, bench_cfg.tag))

            function_input_signature_benchmark_context = hash_sha1(
                (function_input_signature_pure, bench_cfg_sample_hash)
            )
            # logging.info(f"inputs: {fn_inputs_sorted}")
            # logging.info(f"pure: {function_input_signature_pure}")
            if (
                not bench_cfg.overwrite_sample_cache
                and function_input_signature_benchmark_context in self.sample_cache
            ):
                logging.info(
                    f"Hash: {function_input_signature_benchmark_context} was found in context cache, loading..."
                )
                result = self.sample_cache[function_input_signature_benchmark_context]
                self.worker_cache_call_count += 1
            elif (
                not bench_cfg.overwrite_sample_cache
                and bench_run_cfg.only_hash_tag
                and (function_input_signature_pure in self.sample_cache)
            ):
                logging.info(
                    f"Hash: {function_input_signature_benchmark_context} not found in context cache"
                )
                logging.info(
                    f"Hash: {function_input_signature_pure} was found in the function cache, loading..."
                )

                result = self.sample_cache[function_input_signature_pure]
                self.worker_cache_call_count += 1
            else:
                logging.info(f"Context not in cache: {function_input_signature_benchmark_context}")
                logging.info(f"Function inputs not cache: {function_input_signature_pure}")
                logging.info("Calling benchmark function")

                result = self.worker_wrapper(bench_cfg, function_input)
                self.sample_cache.set(
                    function_input_signature_benchmark_context, result, tag=bench_cfg.tag
                )
                self.sample_cache.set(function_input_signature_pure, result, tag=bench_cfg.tag)
        else:
            result = self.worker_wrapper(bench_cfg, function_input)

        # construct a dict for a holomap
        if isinstance(result, dict):  # todo holomaps with named types
            if "hmap" in result:
                # print(isinstance(result["hmap"], hv.element.Element))
                bench_cfg.hmap[canonical_input] = result["hmap"]

        logging.debug(f"input_index {index_tuple}")
        logging.debug(f"input {function_input_vars}")
        logging.debug(f"result {result}")
        for rv in bench_cfg.result_vars:
            if isinstance(result, dict):
                result_value = result[rv.name]
            else:
                result_value = result.param.values()[rv.name]

            if bench_run_cfg.print_bench_results:
                logging.info(f"{rv.name}: {result_value}")

            if isinstance(rv, ResultVar):
                set_xarray_multidim(bench_cfg.ds[rv.name], index_tuple, result_value)
            elif isinstance(rv, ResultVec):
                if isinstance(result_value, (list, np.ndarray)):
                    if len(result_value) == rv.size:
                        for i in range(rv.size):
                            set_xarray_multidim(
                                bench_cfg.ds[rv.index_name(i)], index_tuple, result_value[i]
                            )

            else:
                raise RuntimeError("Unsupported result type")

    def clear_tag_from_cache(self, tag: str):
        """Clear all samples from the cache that match a tag
        Args:
            tag(str): clear samples with this tag
        """
        if self.sample_cache is None:
            self.sample_cache = Cache(
                "cachedir/sample_cache", tag_index=True, size_limit=self.cache_size
            )
        logging.info(f"clearing the sample cache for tag: {tag}")
        removed_vals = self.sample_cache.evict(tag)
        logging.info(f"removed: {removed_vals} items from the cache")

    def add_metadata_to_dataset(self, bench_cfg: BenchCfg, input_var: ParametrizedSweep) -> None:
        """Adds variable metadata to the xrarry so that it can be used to automatically plot the dimension units etc.

        Args:
            bench_cfg (BenchCfg):
            input_var (ParametrizedSweep): The varible to extract metadata from
        """

        for rv in bench_cfg.result_vars:
            if type(rv) == ResultVar:
                bench_cfg.ds[rv.name].attrs["units"] = rv.units
                bench_cfg.ds[rv.name].attrs["long_name"] = rv.name
            elif type(rv) == ResultVec:
                for i in range(rv.size):
                    bench_cfg.ds[rv.index_name(i)].attrs["units"] = rv.units
                    bench_cfg.ds[rv.index_name(i)].attrs["long_name"] = rv.name
            else:
                pass  # todo

        dsvar = bench_cfg.ds[input_var.name]
        dsvar.attrs["long_name"] = input_var.name
        if input_var.units is not None:
            dsvar.attrs["units"] = input_var.units
        if input_var.__doc__ is not None:
            dsvar.attrs["description"] = input_var.__doc__

    def report_results(self, bench_cfg: BenchCfg, print_xarray: bool, print_pandas: bool):
        """Optionally display the caculated benchmark data as either as pandas, xarray or plot

        Args:
            bench_cfg (BenchCfg):
            print_xarray (bool):
            print_pandas (bool):
        """
        if print_xarray:
            logging.info(bench_cfg.ds)
        if print_pandas:
            logging.info(bench_cfg.ds.to_dataframe())

    def clear_call_counts(self) -> None:
        """Clear the worker and cache call counts, to help debug and assert caching is happening properly"""
        self.worker_wrapper_call_count = 0
        self.worker_fn_call_count = 0
        self.worker_cache_call_count = 0

    def get_panel(
        self,
        main_plot: bool = True,
    ) -> pn.pane:
        """Get the panel instance where bencher collates results

        Args:
            main_plot (bool, optional): return the last added page. Defaults to True.

        Returns:
            pn.pane: results panel
        """
        if main_plot:
            if len(self.pane) > 0:
                return self.pane[-1]
            return self.pane
        return self.pane[-1]

    def append_markdown(self, markdown: str, name=None, **kwargs) -> pn.pane.Markdown:
        md = pn.pane.Markdown(markdown, name=name, **kwargs)
        self.append(md, name)
        return md

    def append(self, pane: pn.panel, name: str = None) -> None:
        if name is None:
            name = pane.name
        if len(self.pane) == 0:
            self.append_tab(pane, name)
        else:
            self.pane[-1].append(pane)

    def append_col(self, pane: pn.panel, name=None) -> None:
        if name is not None:
            col = pn.Column(pane, name)
        else:
            col = pn.Column(pane)
        # self.get_panel().append(col)
        self.pane.append(col)

    def append_tab(self, pane: pn.panel, name: str = None) -> None:
        if name is None:
            name = pane.name
        self.pane.append(pn.Column(pane, name=name))

    def save_index(self, directory="", filename="index.html") -> Path:
        """Saves the result to index.html in the root folder so that it can be displayed by github pages.

        Returns:
            Path: save path
        """
        return self.save(directory, filename, False)

    def save(
        self,
        directory: str | Path = "cachedir",
        filename: str = None,
        in_html_folder=True,
        **kwargs,
    ) -> Path:
        """Save the result to a html file.  Note that dynamic content will not work.  by passing save(__file__) the html output will be saved in the same folder as the source code in a html subfolder.

        Args:
            directory (str | Path, optional): base folder to save to. Defaults to "cachedir" which should be ignored by git.
            filename (str, optional): The name of the html file. Defaults to the name of the benchmark
            in_html_folder (bool, optional): Put the saved files in a html subfolder to help keep the results separate from source code. Defaults to True.

        Returns:
            Path: the save path
        """

        if filename is None:
            filename = f"{self.bench_name}.html"

        base_path = Path(directory)

        if in_html_folder:
            base_path /= "html"

        logging.info(f"creating dir {base_path.absolute()}")
        os.makedirs(base_path.absolute(), exist_ok=True)

        base_path = base_path / filename

        logging.info(f"saving html output to: {base_path.absolute()}")

        self.pane.save(filename=base_path, progress=True, **kwargs)
        return base_path

    def publish(self, remote_callback: Callable, branch_name=None, debug: bool = True) -> str:
        """Publish the results as an html file by committing it to the bench_results branch in the current repo. If you have set up your repo with github pages or equivalent then the html file will be served as a viewable webpage.  This is an example of a callable to publish on github pages:

        def publish_args(branch_name) -> Tuple[str, str]:
            return (
                "https://github.com/dyson-ai/bencher.git",
                f"https://github.com/dyson-ai/bencher/blob/{branch_name}",
            )

        Args:
            remote (Callable): A function the returns a tuple of the publishing urls. It must follow the signature def publish_args(branch_name) -> Tuple[str, str].  The first url is the git repo name, the second url needs to match the format for viewable html pages on your git provider.  The second url can use the argument branch_name to point to the report on a specified branch.


        Returns:
            str: the url of the published report
        """

        if branch_name is None:
            branch_name = self.bench_name
        branch_name += "_debug" if debug else ""

        remote, publish_url = remote_callback(branch_name)

        # if debug:
        # publish_url

        directory = "tmpgit"
        report_path = self.save(directory, filename="index.html", in_html_folder=False)
        logging.info(f"created report at: {report_path.absolute()}")

        cd_dir = f"cd {directory} &&"

        os.system(f"{cd_dir} git init")
        os.system(f"{cd_dir} git checkout -b {branch_name}")
        os.system(f"{cd_dir} git add index.html")
        os.system(f'{cd_dir} git commit -m "publish {branch_name}"')
        os.system(f"{cd_dir} git remote add origin {remote}")
        os.system(f"{cd_dir} git push --set-upstream origin {branch_name} -f")

        logging.info("Published report @")
        logging.info(publish_url)

        shutil.rmtree(directory)

        return publish_url

    # def publish_old(
    #     self,
    #     directory: str = "bench_results",
    #     branch_name: str = "bench_results",
    #     url_postprocess: Callable = None,
    #     **kwargs,
    # ) -> str:
    #     """Publish the results as an html file by committing it to the bench_results branch in the current repo. If you have set up your repo with github pages or equivalent then the html file will be served as a viewable webpage.

    #     Args:
    #         directory (str, optional): Directory to save the results. Defaults to "bench_results".
    #         branch_name (str, optional): Branch to publish on. Defaults to "bench_results".
    #         url_postprocess (Callable, optional): A function that maps the origin url to a github pages url. Pass your own function if you are using another git providers. Defaults to None.

    #     Returns:
    #         str: _description_
    #     """

    #     def get_output(cmd: str) -> str:
    #         return (
    #             subprocess.run(cmd.split(" "), stdout=subprocess.PIPE, check=False)
    #             .stdout.decode("utf=8")
    #             .strip()
    #         )

    #     def postprocess_url(publish_url: str, branch_name: str, report_path: str, **kwargs) -> str:
    #         # import re

    #         # return re.sub(
    #         #     """((git|ssh|http(s)?)|(git@[\w\.-]+))(:(//)?)([\w\.@\:/\-~]+)(\.git)(/)?""",
    #         #     """https://$7/""",
    #         #     publish_url,
    #         # )
    #         # git@github.com:user/project.git
    #         # https://github.com/user/project.git
    #         # http://github.com/user/project.git
    #         # git@192.168.101.127:user/project.git
    #         # https://192.168.101.127/user/project.git
    #         # http://192.168.101.127/user/project.git
    #         # ssh://user@host.xz:port/path/to/repo.git/
    #         # ssh://user@host.xz/path/to/repo.git/
    #         # ssh://host.xz:port/path/to/repo.git/
    #         # ssh://host.xz/path/to/repo.git/
    #         # ssh://user@host.xz/path/to/repo.git/
    #         # ssh://host.xz/path/to/repo.git/
    #         # ssh://user@host.xz/~user/path/to/repo.git/
    #         # ssh://host.xz/~user/path/to/repo.git/
    #         # ssh://user@host.xz/~/path/to/repo.git
    #         # ssh://host.xz/~/path/to/repo.git
    #         # git://host.xz/path/to/repo.git/
    #         # git://host.xz/~user/path/to/repo.git/
    #         # http://host.xz/path/to/repo.git/
    #         # https://host.xz/path/to/repo.git/
    #         # https://regex101.com/r/qT7NP0/3

    #         return publish_url.replace(".git", f"/blob/{directory}/{report_path}")

    #     if url_postprocess is None:
    #         url_postprocess = postprocess_url
    #     current_branch = get_output("git symbolic-ref --short HEAD")
    #     logging.info(f"on branch: {current_branch}")
    #     stash_msg = get_output("git stash")
    #     logging.info(f"stashing current work :{stash_msg}")
    #     checkout_msg = get_output(f"git checkout -b {branch_name}")
    #     checkout_msg = get_output(f"git checkout {branch_name}")
    #     get_output("git pull")

    #     logging.info(f"checking out branch: {checkout_msg}")
    #     report_path = self.save(directory, in_html_folder=False)
    #     logging.info(f"created report at: {report_path.absolute()}")
    #     # commit_msg = f""
    #     logging.info("adding report to git")
    #     get_output(f"git add {report_path.absolute()}")
    #     get_output("git status")
    #     logging.info("committing report")
    #     cmd = f'git commit -m "generate_report:{self.bench_name}"'
    #     logging.info(cmd)
    #     get_output(cmd)
    #     logging.info("pushing report to origin")
    #     get_output(f"git push --set-upstream origin {branch_name}")
    #     logging.info("checking out original branch")
    #     get_output(f"git checkout {current_branch}")
    #     if "No local changes" not in stash_msg:
    #         logging.info("restoring work with git stash pop")
    #         get_output("git stash pop")

    #     publish_url = get_output("git remote get-url --push origin")
    #     logging.info(f"raw url:{publish_url}")
    #     publish_url = url_postprocess(
    #         publish_url, branch_name=branch_name, report_path=report_path, **kwargs
    #     )
    #     logging.info("Published report @")
    #     logging.info(publish_url)
    #     return publish_url



# from typing import Optional
# import matplotlib.pyplot as plt
# import panel as pn
# import seaborn as sns

# from bencher.plotting.plot_filter import PlotFilter, PlotInput, VarRange
# from bencher.plotting.plot_types import PlotTypes
# from bencher.plotting_functions import wrap_long_time_labels


# class Scatter:
#     """A class to display the result data as a scatter plot"""

#     def scatter2D_sns(self, pl_in: PlotInput) -> Optional[pn.panel]:
#         """Given a benchCfg generate a 2D scatter plot from seaborn

#         Args:
#             pl_in (PlotInput): data to plot

#         Returns:
#             Optional[pn.pane]: A panel pane with a scatter plot
#         """

#         if PlotFilter(
#             float_range=VarRange(None, None),
#             cat_range=VarRange(None, None),
#             vector_len=VarRange(2, 2),
#         ).matches(pl_in.plt_cnt_cfg):
#             bench_cfg = wrap_long_time_labels(pl_in.bench_cfg)
#             ds = bench_cfg.ds.drop_vars("repeat")

#             df = ds.to_dataframe().reset_index()

#             names = pl_in.rv.index_names()

#             if bench_cfg.input_vars:
#                 h = sns.jointplot(df, x=names[0], y=names[1], hue=bench_cfg.input_vars[0].name)
#             elif bench_cfg.over_time:
#                 h = sns.jointplot(df, x=names[0], y=names[1], hue=bench_cfg.iv_time[0].name)

#             else:
#                 h = sns.jointplot(df, x=names[0], y=names[1])

#             h.set_axis_labels(f"{names[0]} [{pl_in.rv.units}]", f"{names[1]} [{pl_in.rv.units}]")
#             plt.tight_layout()
#             return pn.panel(h.fig, name=PlotTypes.scatter2D_sns)
#         return None

#     # def plot_scatter3D_px(self, pl_in: PlotInput) -> pn.pane.Plotly:
#     #     """Given a benchCfg generate a 3D scatter plot with plotly express

#     #     Args:
#     #         bench_cfg (BenchCfg): description of benchmark
#     #         rv (ParametrizedSweep): result variable to plot

#     #     Returns:
#     #         pn.pane.Plotly: A 3d scatter plot as a holoview in a pane
#     #     """

#     #     bench_cfg = pl_in.bench_cfg
#     #     rv = pl_in.rv

#     #     bench_cfg = wrap_long_time_labels(bench_cfg)

#     #     df = bench_cfg.get_dataframe()

#     #     names = rv.index_names()  # get the column names of the vector result

#     #     if bench_cfg.input_vars:
#     #         color = bench_cfg.input_vars[0].name
#     #     else:
#     #         color = None

#     #     if bench_cfg.over_time:
#     #         if len(names) < 3:  # only a 2d vector result so use the time axis as the third point
#     #             names.insert(0, bench_cfg.iv_time[0].name)

#     #     return px.scatter_3d(
#     #         df, x=names[0], y=names[1], z=names[2], color=color, symbol=color, size_max=2
#     #     )



import unittest
import bencher as bch
from bencher.example.example_categorical import example_categorical
from bencher.example.example_floats import example_floats
from bencher.example.example_floats2D import example_floats2D
from bencher.example.example_pareto import example_pareto
from bencher.example.example_simple_cat import example_1D_cat
from bencher.example.example_simple_float import example_1D_float
from bencher.example.example_float_cat import run_example_float_cat
from bencher.example.example_time_event import run_example_time_event
from bencher.example.example_float3D import example_floats3D

from bencher.example.example_custom_sweep import example_custom_sweep
from bencher.example.example_workflow import example_floats2D_workflow, example_floats3D_workflow
from bencher.example.example_plot_library import example_plot_library
from bencher.example.example_holosweep_tap import example_holosweep_tap

from bencher.example.optuna.example_optuna import optuna_rastrigin
from bencher.example.example_sample_cache import example_sample_cache

import os

# shelved
# from bencher.example.shelved.example_float2D_scatter import example_floats2D_scatter
# from bencher.example.shelved.example_float3D_cone import example_cone


class TestBenchExamples(unittest.TestCase):
    """The purpose of this test class is to run the example problems to make sure they don't crash.  The bencher logic is tested in the other test files test_bencher.py and test_vars.py"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_all = False

    def create_run_cfg(self) -> bch.BenchRunCfg:
        cfg = bch.BenchRunCfg()
        if not self.generate_all:
            cfg.repeats = 2  # low number of repeats to reduce test time, but also test averaging and variance code
            cfg.debug = True  # reduce size of param sweep so tests are faster
        cfg.clear_cache = True
        return cfg

    def examples_asserts(self, example_result, save=False) -> None:
        self.assertIsNotNone(example_result)
        if save or self.generate_all:
            # example_result.save("bencher/example/html")
            path = example_result.save_index("cachedir")
            self.assertTrue(os.path.exists(path))

    def test_example_categorical(self) -> None:
        self.examples_asserts(example_categorical(self.create_run_cfg()))

    def test_example_floats(self) -> None:
        self.examples_asserts(example_floats(self.create_run_cfg()))

    def test_example_floats2D(self) -> None:
        self.examples_asserts(example_floats2D(self.create_run_cfg()))

    def test_example_pareto(self) -> None:
        self.examples_asserts(example_pareto(self.create_run_cfg()))

    def test_example_simple_cat(self) -> None:
        self.examples_asserts(example_1D_cat(self.create_run_cfg()), save=True)

    def test_example_simple_float(self) -> None:
        self.examples_asserts(example_1D_float(self.create_run_cfg()))

    def test_example_float_cat(self) -> None:
        self.examples_asserts(run_example_float_cat(self.create_run_cfg()))

    def test_example_time_event(self) -> None:
        self.examples_asserts(run_example_time_event(self.create_run_cfg()))

    def test_example_float3D(self) -> None:
        self.examples_asserts(example_floats3D(self.create_run_cfg()))

    def test_example_custom_sweep(self) -> None:
        self.examples_asserts(example_custom_sweep(self.create_run_cfg()))

    def test_example_floats2D_workflow(self) -> None:
        self.examples_asserts(example_floats2D_workflow(self.create_run_cfg()))

    def test_example_floats3D_workflow(self) -> None:
        self.examples_asserts(example_floats3D_workflow(self.create_run_cfg()))

    def test_plot_library(self) -> None:
        self.examples_asserts(example_plot_library(self.create_run_cfg()))

    def test_holosweep_tap(self) -> None:
        self.examples_asserts(example_holosweep_tap(self.create_run_cfg()))

    def test_optuna_rastrigin(self) -> None:
        self.examples_asserts(optuna_rastrigin(self.create_run_cfg()))

    def test_example_sample_cache(self) -> None:
        self.examples_asserts(example_sample_cache(self.create_run_cfg(), False))

    # shelved
    # def test_example_cone(self) -> None:
    #     self.examples_asserts(example_cone(self.create_run_cfg()))
    # def test_float2D_scatter(self) -> None:
    #     self.examples_asserts(example_floats2D_scatter(self.create_run_cfg()))



from __future__ import annotations

import inspect
import logging
from typing import Callable, List

import panel as pn

from bencher.bench_cfg import BenchCfg, PltCntCfg
from bencher.plotting.plot_filter import PlotInput, PlotProvider
from bencher.variables.parametrised_sweep import ParametrizedSweep


class PlotCollection:
    """A set of plot providers with filters that determine if they are able to plot the type of data in the results.  The plot collection can be customised with user defined plots or ones already defined by bencher. You can also take predefine PlotCollections defined in PlotLibrary and add or remove specific plots that you want to override.  The add and remove function accept strings, or StrEnums (which are equivalent to strings)"""

    def __init__(self) -> None:
        """A PlotCollection has a dictionary of function_name:function_pointer which are all the available types of plot. The plotters variable contains an ordered dictionary of the active plotting functions"""
        self.plotter_providers = {}
        self.plotters = {}

    def add_plotter_source(self, plotter_class_instance: PlotProvider) -> None:
        """Given a class definition that contains several plotting methods, get a list of all the methods of the class and add them to the dictionary of available plotting functions

        Args:
            plotter_class_instance (PlotProvider): A class that contains plotting methods
        """

        if not inspect.isclass(plotter_class_instance):
            self.plotter_providers |= dict(
                inspect.getmembers(plotter_class_instance, predicate=inspect.ismethod)
            )
        else:
            raise ValueError(
                "You need to pass an instance of a class not the class type, Add () to the end of the class name"
            )

    def add(self, plot_name: str, plot_fn: Callable = None) -> PlotCollection:
        """Add a plotting method to the list of active plotting functions

        Args:
            plot_name (str): The name of the plot to add.  This can either be a string, or from the StrEnum PlotTypes (which is also a string)
            plot_fn (Callable): A user defined plotting function.  If no function is supplied, the name of the function is looked up from the list of plotter sources that have been added to the plot collection.

        Returns:
            PlotCollection: Returns a reference to this class so that you can call with a fluent api, e.g plot_coll.add("barplot").remove("swarmplot")
        """
        if plot_fn is None:
            if plot_name in self.plotter_providers:
                self.plotters[plot_name] = self.plotter_providers[plot_name]
            else:
                logging.warning(f"{plot_name} plot was not found in the list of available plots")
                raise ValueError(f"{plot_name} plot was not found in the list of available plots")
        else:
            self.plotters[plot_name] = plot_fn
        return self

    def add_list(self, plot_names: List[str]) -> PlotCollection:
        """Add a list of plotting methods to the list of active plotting functions

        Args:
            plot_name (List[str]): A list of names of the plot to add. The list can contain either strings, or StrEnum PlotTypes (which is also a string)

        Returns:
            PlotCollection: Returns a reference to this class so that you can call with a fluent api, e.g plot_coll.add_list(["barplot","boxplot"]).remove("swarmplot")
        """
        for p in plot_names:
            self.add(p)
        return self

    def remove(self, plot_name: str) -> PlotCollection:
        """Remove a plotting method from the list of active plotting functions

        Args:
            plot_name (str): The name of the plot to add.  This can either be a string, or from the StrEnum PlotTypes (which is also a string)

        Returns:
            PlotCollection: Returns a reference to this class so that you can call with a fluent api, e.g plot_coll.remove(swarmplot).add(barplot)
        """
        self.plotters.pop(plot_name)
        return self

    def gather_plots(
        self, bench_cfg: BenchCfg, rv: ParametrizedSweep, plt_cnt_cfg: PltCntCfg
    ) -> pn.pane.panel:
        """This method returns a single plot based on 1 result variable and a set of input variables.  It dedeuces the correct plot type by passing it to several configuration functions that operate on the number of inputs

        Args:
            bench_cfg (BenchCfg): A config of the input vars
            rv (ParametrizedSweep): a config of the result variable
            plt_cnt_cfg (PltCntCfg): A config of how many input types there are
        Raises:
            ValueError: If the plot does not inherit from a pn.viewable.Viewable type
        """

        tabs = pn.Accordion()
        for plt_fn in self.plotters.values():
            plots = plt_fn(PlotInput(bench_cfg, rv, plt_cnt_cfg))
            if plots is not None:
                if not isinstance(plots, list):
                    plots = [plots]
                for plt_instance in plots:
                    if not isinstance(plt_instance, pn.viewable.Viewable):
                        raise ValueError(
                            "The plot must be a viewable type (pn.viewable.Viewable or pn.panel)"
                        )
                    logging.info(f"plotting: {plt_instance.name}")
                    tabs.append(plt_instance)
        if len(tabs) > 0:
            tabs.active = list(range(len(tabs)))  # set all plots as active
        return tabs



from typing import Optional

import panel as pn
import seaborn as sns
import hvplot.xarray  # noqa pylint: disable=unused-import

from bencher.plotting.plot_filter import PlotFilter, PlotInput, VarRange
from bencher.plotting.plot_types import PlotTypes
from bencher.plotting.plots.catplot import Catplot
from bencher.plt_cfg import PltCfgBase
from bencher.plotting.plots.plot_base import PlotBase


class Lineplot(PlotBase):
    # shared plot filter for lineplots
    plot_filter = PlotFilter(
        float_range=VarRange(1, 1),
        cat_range=VarRange(0, 1),
        vector_len=VarRange(1, 1),
        result_vars=VarRange(1, 1),
    )

    def lineplot(self, pl_in: PlotInput) -> Optional[pn.panel]:
        """generate a line plot

        Args:
            pl_in (PlotInput): data to plot

        Returns:
            Optional[pn.panel]: a line plot of the data
        """
        if self.plot_filter.matches(pl_in.plt_cnt_cfg):
            df = pl_in.bench_cfg.ds[pl_in.rv.name].to_dataframe().reset_index()
            sns_cfg = PltCfgBase()

            x = pl_in.plt_cnt_cfg.float_vars[0]
            y = pl_in.rv
            if pl_in.plt_cnt_cfg.cat_cnt > 0:
                sns_cfg.hue = pl_in.plt_cnt_cfg.cat_vars[0].name

            sns_cfg.x = x.name
            sns_cfg.y = y.name

            sns_cfg.kind = "line"
            sns_cfg.xlabel = f"{x.name} [{x.units}]"
            sns_cfg.ylabel = f"{sns_cfg.y} [{pl_in.rv.units}]"
            sns_cfg.title = self.title(x, y)

            fg = sns.relplot(df, **sns_cfg.as_sns_args())

            # TODO try to set this during the initial plot rather than after
            for ax in fg.axes.flatten():
                for tick in ax.get_xticklabels():
                    tick.set_rotation(45)

            return Catplot.plot_postprocess(fg, sns_cfg, PlotTypes.lineplot)
        return None

    # def lineplot_hv(self, pl_in: PlotInput) -> Optional[pn.panel]:
    #     """generate a line plot

    #     Args:
    #         pl_in (PlotInput): data to plot

    #     Returns:
    #         Optional[pn.panel]: a line plot of the data
    #     """
    #     if self.plot_filter.matches(pl_in.plt_cnt_cfg):
    #         da = pl_in.bench_cfg.ds[pl_in.rv.name]

    #         x = pl_in.plt_cnt_cfg.float_vars[0]
    #         y = pl_in.rv
    #         title = self.title(x, y)

    #         col = None
    #         if pl_in.plt_cnt_cfg.cat_cnt > 0:
    #             col = pl_in.plt_cnt_cfg.cat_vars[0]

    #         da[col.name] = ["absolute", "negate"]

    #         kwargs = {"x": x.name, "y": y.name}

    #         print(kwargs)

    #         pt = hv.Overlay()

    #         # look into:
    #         # https://holoviews.org/user_guide/Dimensioned_Containers.html

    #         df = da.to_dataframe()
    #         df = df.pivot_table(columns=[col.name], index=[x.name], values=[y.name])[y.name]
    #         print(df)
    #         # dag = da.groupby(col.name)[y.name]
    #         # dag = da.groupby(col.name)
    #         pt = df.hvplot.line()
    #         # pt *= dag.hvplot.line()
    #         # pt *= df.hvplot.line(y=col.values(False))
    #         # pt *= da.hvplot.line(groupby=col.name, **kwargs)
    #         # pt.opts(shared_axes=False, title=title)
    #         pt.opts(title=title)

    #         # mean = da.mean("repeat", keepdims=True, keep_attrs=True)
    #         return pn.Column(pt, name=PlotTypes.lineplot_hv)
    #     return None

    # def lineplot_hv_repeats(self, pl_in: PlotInput) -> Optional[pn.panel]:
    #     """generate a line plot

    #     Args:
    #         pl_in (PlotInput): data to plot

    #     Returns:
    #         Optional[pn.panel]: a line plot of the data
    #     """
    #     if self.plot_filter.matches(pl_in.plt_cnt_cfg):
    #         da = pl_in.bench_cfg.ds[pl_in.rv.name]

    #         x = pl_in.plt_cnt_cfg.float_vars[0]
    #         y = pl_in.rv
    #         title = self.title(x, y)

    #         col = None
    #         if pl_in.plt_cnt_cfg.cat_cnt > 0:
    #             col = pl_in.plt_cnt_cfg.cat_vars[0].name

    #         pt = hv.Overlay()
    #         hvds = hv.Dataset(da)
    #         import numpy as np

    #         # to read
    #         # https://medium.com/@shouke.wei/data-visualization-with-hvplot-iii-interactive-multiple-plots-e73124742df4

    #         # pt *= hvds.to(hv.Curve, x.name, "repeat")
    #         # pt *= hv.Curve(hvds, kdims=[x.name], vdims=[y.name])
    #         # pt *= hv.Spread(hvds.aggregate(y.name, np.mean, np.std), kdims=[y.name])
    #         # return None
    #         print(hvds)

    #         pt += hvds

    #         pt *= hv.Spread(
    #             hvds.aggregate("repeat", np.mean, np.std), kdims=[x.name], vdims=[y.name]
    #         )
    #         pt *= hv.Curve(hvds.aggregate("repeat", np.mean), kdims=[x.name], vdims=[y.name])

    #         col = pn.Column(name="spread")

    #         col.append(pt)

    #         return col
    #         # return da.to(hv.Spread)

    #         stats, std = self.calculate_stats(da)

    #         kwargs = {"x": x.name, "y": "mean"}

    #         std_kwargs = kwargs | {"y": "std_low", "y2": "std_high"}

    #         pt *= stats.hvplot.area(**std_kwargs).opts(alpha=0.2)
    #         # pt *= hv.Spread(std).opts(alpha=0.2)
    #         pt *= stats.hvplot.line(**kwargs)

    #         pt.opts(title=title)

    #         return pn.Column(pt, name=PlotTypes.lineplot_hv_repeats)
    #     return None

    # def lineplot_hv_subplot(self, pl_in: PlotInput) -> Optional[pn.panel]:
    #     """generate a line plot

    #     Args:
    #         pl_in (PlotInput): data to plot

    #     Returns:
    #         Optional[pn.panel]: a line plot of the data
    #     """
    #     if PlotFilter(
    #         float_range=VarRange(1, 1),
    #         cat_range=VarRange(1, 1),
    #         vector_len=VarRange(1, 1),
    #         result_vars=VarRange(1, 1),
    #     ).matches(pl_in.plt_cnt_cfg):
    #         da = pl_in.bench_cfg.ds[pl_in.rv.name]
    #         da = da.mean("repeat")

    #         lines = pl_in.plt_cnt_cfg.cat_vars[0].values(False)
    #         print(lines)
    #         return pn.Column(
    #             da.hvplot.line(subplots=True),
    #             name=PlotTypes.lineplot_hv_subplot,
    #         )
    #     return None



import unittest
from hypothesis import given, strategies as st  # pylint: disable=unused-import
import pytest
from bencher.variables.inputs import IntSweep, EnumSweep, StringSweep, BoolSweep, FloatSweep
from enum import auto
from strenum import StrEnum


class TestVarSweeps(unittest.TestCase):
    def test_int_sweep_01(self):
        int_sweep = IntSweep(bounds=[0, 1])
        self.assertEqual(int_sweep.default, 0)
        self.assertListEqual(int_sweep.values(False), [0, 1])
        self.assertListEqual(int_sweep.values(True), [0, 1])

    def test_int_sweep_06(self):
        int_sweep = IntSweep(bounds=[0, 6])
        self.assertEqual(int_sweep.default, 0)
        self.assertListEqual(int_sweep.values(False), [0, 1, 2, 3, 4, 5, 6])
        self.assertListEqual(int_sweep.values(True), [0, 6])

    def test_int_sweep_06_debug_sampes(self):
        int_sweep = IntSweep(bounds=[0, 6], samples_debug=4)
        self.assertEqual(int_sweep.default, 0)
        self.assertListEqual(int_sweep.values(False), [0, 1, 2, 3, 4, 5, 6])
        self.assertListEqual(int_sweep.values(True), [0, 2, 4, 6])

    def test_int_sweep_10_debug_sampes(self):
        int_sweep = IntSweep(bounds=[0, 10])
        self.assertEqual(int_sweep.default, 0)
        self.assertListEqual(int_sweep.values(False), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.assertListEqual(int_sweep.values(True), [0, 10])

    def test_int_sweep_10_with_values_(self):
        int_sweep = IntSweep(bounds=[0, 10])
        self.assertEqual(int_sweep.default, 0)
        self.assertListEqual(int_sweep.values(False), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.assertListEqual(int_sweep.with_samples(3).values(True), [0, 10])
        self.assertListEqual(int_sweep.with_samples(3).values(False), [0, 5, 10])

    def test_int_sweep_level_10_with_values_(self):
        int_sweep = IntSweep(bounds=[0, 10])
        self.assertListEqual(int_sweep.with_level(1).values(), [0])
        self.assertListEqual(int_sweep.with_level(2).values(), [0, 10])
        self.assertListEqual(int_sweep.with_level(3).values(), [0, 5, 10])
        self.assertListEqual(int_sweep.with_level(4).values(), [0, 2, 5, 7, 10])
        self.assertListEqual(int_sweep.with_level(5).values(), [0, 1, 2, 3, 5, 6, 7, 8, 10])
        self.assertListEqual(int_sweep.with_level(6).values(), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def test_float_sweep(self):
        float_sweep = FloatSweep(bounds=[0, 1])

        self.assertListEqual(list(float_sweep.with_level(1).values()), [0.0])
        self.assertListEqual(list(float_sweep.with_level(2).values()), [0.0, 1.0])
        self.assertListEqual(list(float_sweep.with_level(3).values()), [0.0, 0.5, 1.0])
        self.assertListEqual(list(float_sweep.with_level(4).values()), [0.0, 0.25, 0.5, 0.75, 1.0])

    def test_float_step(self):
        float_sweep = FloatSweep(bounds=[0, 1], step=0.1)

        self.assertListEqual(
            list(float_sweep.values()),
            [
                0.0,
                0.1,
                0.2,
                0.30000000000000004,
                0.4,
                0.5,
                0.6000000000000001,
                0.7000000000000001,
                0.8,
                0.9,
            ],
        )

        float_sweep2 = FloatSweep(bounds=[0, 0.9], step=0.1)
        # note that it does not return the upper bound of 0.9 even though it a multiple of 0.1
        self.assertListEqual(
            list(float_sweep2.values()),
            [
                0.0,
                0.1,
                0.2,
                0.30000000000000004,
                0.4,
                0.5,
                0.6000000000000001,
                0.7000000000000001,
                0.8,
            ],
        )

    def test_enum_sweep_level(self):
        class Enum1(StrEnum):
            VAL1 = auto()
            VAL2 = auto()
            VAL3 = auto()
            VAL4 = auto()

        enum_sweep = EnumSweep(Enum1)

        self.assertListEqual(enum_sweep.with_level(1).values(), [Enum1.VAL1])

        self.assertListEqual(enum_sweep.with_level(2).values(), [Enum1.VAL1, Enum1.VAL4])

        self.assertListEqual(
            enum_sweep.with_level(3).values(), [Enum1.VAL1, Enum1.VAL2, Enum1.VAL4]
        )
        self.assertListEqual(
            enum_sweep.with_level(4).values(), [Enum1.VAL1, Enum1.VAL2, Enum1.VAL3, Enum1.VAL4]
        )
        self.assertListEqual(
            enum_sweep.with_level(5).values(), [Enum1.VAL1, Enum1.VAL2, Enum1.VAL3, Enum1.VAL4]
        )

    def test_bool_default(self) -> None:
        bool_sweep_true = BoolSweep(default=True)
        self.assertTrue(bool_sweep_true.default)

        bool_sweep_false = BoolSweep(default=False)
        self.assertFalse(bool_sweep_false.default)

    def test_bool_sweep_level(self):
        bool_sweep = BoolSweep()

        self.assertListEqual(bool_sweep.with_level(1).values(), [True])
        self.assertListEqual(bool_sweep.with_level(2).values(), [True, False])
        self.assertListEqual(bool_sweep.with_level(3).values(), [True, False])

    def test_string_sweep_level(self):
        string_sweep = StringSweep(["VAL1", "VAL2", "VAL3", "VAL4"])
        self.assertListEqual(string_sweep.with_level(1).values(), ["VAL1"])
        self.assertListEqual(string_sweep.with_level(2).values(), ["VAL1", "VAL4"])
        self.assertListEqual(string_sweep.with_level(3).values(), ["VAL1", "VAL2", "VAL4"])
        self.assertListEqual(string_sweep.with_level(4).values(), ["VAL1", "VAL2", "VAL3", "VAL4"])
        self.assertListEqual(string_sweep.with_level(5).values(), ["VAL1", "VAL2", "VAL3", "VAL4"])

    @given(st.integers(min_value=1, max_value=10))
    def test_int_sweep_samples(self, samples):
        int_sweep = IntSweep(bounds=[0, 10], samples=samples)
        self.assertEqual(int_sweep.default, 0)
        self.assertEqual(len(int_sweep.values(False)), samples)

    @given(st.integers(min_value=1, max_value=10))
    def test_int_sweep_debug_samples(self, samples_debug):
        int_sweep = IntSweep(bounds=[0, 10], samples_debug=samples_debug)
        self.assertEqual(int_sweep.default, 0)
        self.assertEqual(len(int_sweep.values(True)), samples_debug)

    @pytest.mark.skip
    @given(st.integers(min_value=1, max_value=10), st.integers(min_value=1, max_value=10))
    def test_int_sweep_samples_all(self, samples, samples_debug):
        int_sweep = IntSweep(bounds=[0, 10], samples=samples)
        self.assertEqual(int_sweep.default, 0)
        self.assertEqual(len(int_sweep.values(False)), samples)
        self.assertEqual(len(int_sweep.values(True)), samples_debug)



from __future__ import annotations

import argparse
import copy
import logging
from typing import Any, Callable, List, Tuple

import param
import xarray as xr
from pandas import DataFrame
from str2bool import str2bool
import holoviews as hv
import numpy as np
import panel as pn


import bencher as bch
from bencher.variables.sweep_base import hash_sha1, describe_variable
from bencher.variables.time import TimeSnapshot, TimeEvent
from bencher.variables.results import OptDir
from bencher.utils import hmap_canonical_input

from enum import Enum, auto


class ReduceType(Enum):
    AUTO = auto()
    SQUEEZE = auto()
    REDUCE = auto()
    NONE = auto()


class PltCfgBase(param.Parameterized):
    """A base class that contains plotting parameters shared by seaborn and xarray"""

    x: str = param.String(None, doc="the x parameter for seaborn and xarray plotting functions")
    y: str = param.String(None, doc="the y parameter for seaborn and xarray plotting functions")
    z: str = param.String(None, doc="the z parameter for xarray plotting functions")
    hue: str = param.String(None, doc="the hue parameter for seaborn and xarray plotting functions")

    row: str = param.String(None, doc="the row parameter for seaborn and xarray plotting functions")
    col: str = param.String(None, doc="the col parameter for seaborn and xarray plotting functions")

    num_rows: int = param.Integer(1, doc="The number of rows in the facetgrid")
    num_cols: int = param.Integer(1, doc="The number of cols in the facetgrid")

    kind: str = param.String(None, doc="the 'kind' of plot, ie barplot, lineplot, scatter etc")
    marker: str = param.String(None, doc="The marker to use when plotting")

    cmap = param.String(None, doc="colormap type")

    plot_callback_sns: Callable = None
    plot_callback_xra: Callable = None
    plot_callback: Callable = None

    xlabel: str = param.String(None, doc="The label for the x-axis")
    ylabel: str = param.String(None, doc="The label for the y-axis")
    zlabel: str = param.String(None, doc="The label for the z-axis")

    title: str = param.String(None, doc="The title of the graph")

    def as_dict(self, include_params: List[str] = None, exclude_params: List[str] = None) -> dict:
        """Return this class as dictionary to pass to plotting functions but exclude parameters that are not expected by those functions

        Args:
            include_params (List[str], optional): a list of property names to include. Defaults to None.
            exclude_params (List[str], optional): a list of property names to exclude. Defaults to None.

        Returns:
            dict: a dictionary of plotting arguments
        """
        all_params = self.param.values()
        all_params.pop("name")

        output = {}
        if include_params is not None:
            for i in include_params:
                output[i] = all_params[i]
        else:
            output = all_params

        if exclude_params is not None:
            for e in exclude_params:
                output.pop(e)

        return output

    def as_sns_args(self) -> dict:
        return self.as_dict(include_params=["x", "y", "hue", "row", "col", "kind"])


class PltCntCfg(param.Parameterized):
    """Plot Count Config"""

    float_vars = param.List(doc="A list of float vars in order of plotting, x then y")
    float_cnt = param.Integer(0, doc="The number of float variables to plot")
    cat_vars = param.List(doc="A list of categorical values to plot in order hue,row,col")
    cat_cnt = param.Integer(0, doc="The number of cat variables")
    vector_len = param.Integer(1, doc="The vector length of the return variable , scalars = len 1")
    result_vars = param.Integer(1, doc="The number result variables to plot")


class BenchPlotSrvCfg(param.Parameterized):
    port: int = param.Integer(None, doc="The port to launch panel with")
    allow_ws_origin = param.Boolean(
        False,
        doc="Add the port to the whilelist, (warning will disable remote access if set to true)",
    )


class BenchRunCfg(BenchPlotSrvCfg):
    """A Class to store options for how to run a benchmark parameter sweep"""

    repeats: bool = param.Integer(1, doc="The number of times to sample the inputs")

    over_time: bool = param.Boolean(
        False,
        doc="If true each time the function is called it will plot a timeseries of historical and the latest result.",
    )

    debug: bool = param.Boolean(
        False, doc="Debug the sampling faster by reducing the dimension sampling resolution"
    )

    use_optuna: bool = param.Boolean(False, doc="show optuna plots")

    summarise_constant_inputs = param.Boolean(
        True, doc="Print the inputs that are kept constant when describing the sweep parameters"
    )

    print_bench_inputs: bool = param.Boolean(
        True, doc="Print the inputs to the benchmark function every time it is called"
    )

    print_bench_results: bool = param.Boolean(
        True, doc="Print the results of the benchmark function every time it is called"
    )

    clear_history: bool = param.Boolean(False, doc="Clear historical results")

    print_pandas: bool = param.Boolean(
        False, doc="Print a pandas summary of the results to the console."
    )

    print_xarray: bool = param.Boolean(
        False, doc="Print an xarray summary of the results to the console"
    )

    serve_pandas: bool = param.Boolean(
        False,
        doc="Serve a pandas summary on the results webpage.  If you have a large dataset consider setting this to false if the page loading is slow",
    )

    serve_pandas_flat: bool = param.Boolean(
        True,
        doc="Serve a flattend pandas summary on the results webpage.  If you have a large dataset consider setting this to false if the page loading is slow",
    )

    serve_xarray: bool = param.Boolean(
        False,
        doc="Serve an xarray summary on the results webpage. If you have a large dataset consider setting this to false if the page loading is slow",
    )

    auto_plot: bool = param.Boolean(
        True, doc=" Automaticlly dedeuce the best type of plot for the results."
    )

    raise_duplicate_exception: bool = param.Boolean(False, doc=" Used to debug unique plot names.")

    use_cache: bool = param.Boolean(
        False,
        doc="This is a benchmark level cache that stores the results of a fully completed benchmark. At the end of a benchmark the values are added to the cache but are not if the benchmark does not complete.  If you want to cache values during the benchmark you need to use the use_sample_cache option. Beware that depending on how you change code in the objective function, the cache could provide values that are not correct.",
    )

    clear_cache: bool = param.Boolean(
        False, doc=" Clear the cache of saved input->output mappings."
    )

    use_sample_cache: bool = param.Boolean(
        False,
        doc="If true, every time the benchmark function is called, bencher will check if that value has been calculated before and if so load the from the cache.  Note that the sample level cache is different from the benchmark level cache which only caches the aggregate of all the results at the end of the benchmark. This cache lets you stop a benchmark halfway through and continue. However, beware that depending on how you change code in the objective function, the cache could provide values that are not correct.",
    )

    only_hash_tag: bool = param.Boolean(
        False,
        doc="By default when checking if a sample has been calculated before it includes the hash of the greater benchmarking context.  This is safer because it means that data generated from one benchmark will not affect data from another benchmark.  However, if you are careful it can be more flexible to ignore which benchmark generated the data and only use the tag hash to check if that data has been calculated before. ie, you can create two benchmarks that sample a subset of the problem during exploration and give them the same tag, and then afterwards create a larger benchmark that covers the cases you already explored.  If this value is true, the combined benchmark will use any data from other benchmarks with the same tag.",
    )

    clear_sample_cache: bool = param.Boolean(
        False,
        doc="Clears the per-sample cache.  Use this if you get unexpected behavior.  The per_sample cache is tagged by the specific benchmark it was sampled from. So clearing the cache of one benchmark will not clear the cache of other benchmarks.",
    )

    overwrite_sample_cache: bool = param.Boolean(
        False,
        doc="If True, recalculate the value and overwrite the value stored in the sample cache",
    )

    only_plot: bool = param.Boolean(
        False, doc="Do not attempt to calculate benchmarks if no results are found in the cache"
    )

    use_holoview: bool = param.Boolean(False, doc="Use holoview for plotting")

    nightly: bool = param.Boolean(
        False, doc="Run a more extensive set of tests for a nightly benchmark"
    )

    time_event: str = param.String(
        None,
        doc="A string representation of a sequence over time, i.e. datetime, pull request number, or run number",
    )

    headless: bool = param.Boolean(False, doc="Run the benchmarks headlessly")

    render_plotly = param.Boolean(
        True,
        doc="Plotly and Bokeh don't play nicely together, so by default pre-render plotly figures to a non dynamic version so that bokeh plots correctly.  If you want interactive 3D graphs, set this to true but be aware that your 2D interactive graphs will probalby stop working.",
    )

    level = param.Integer(
        default=0,
        bounds=[0, 12],
        doc="The level parameter is a method of defining the number samples to sweep over in a variable agnostic way, i.e you don't need to specficy the number of samples for each variable as they are calculated dynamically from the sampling level.  See example_level.py for more information.",
    )

    run_tag = param.String(
        default="",
        doc="Define a tag for a run to isolate the results stored in the cache from other runs",
    )

    @staticmethod
    def from_cmd_line() -> BenchRunCfg:
        """create a BenchRunCfg by parsing command line arguments

        Returns:
            parsed args: parsed args
        """

        parser = argparse.ArgumentParser(description="benchmark")

        parser.add_argument(
            "--use-cache",
            action="store_true",
            help=BenchRunCfg.param.use_cache.doc,
        )

        parser.add_argument(
            "--only-plot",
            action="store_true",
            help=BenchRunCfg.param.only_plot.doc,
        )

        parser.add_argument(
            "--port",
            type=int,
            help=BenchRunCfg.param.port.doc,
        )

        parser.add_argument(
            "--nightly",
            type=lambda b: bool(str2bool(b)),
            nargs="?",
            const=False,
            default=False,
            help="turn on nightly benchmarking",
        )

        parser.add_argument(
            "--time_event",
            type=str,
            default=BenchRunCfg.param.time_event.default,
            help=BenchRunCfg.param.time_event.doc,
        )

        parser.add_argument(
            "--repeats",
            type=int,
            default=BenchRunCfg.param.repeats.default,
            help=BenchRunCfg.param.repeats.doc,
        )

        return BenchRunCfg(**vars(parser.parse_args()))


class BenchCfg(BenchRunCfg):
    """A class for storing the arguments to configure a benchmark protocol  If the inputs variables are the same the class should return the same hash and same filename.  This is so that historical data can be referenced and ensures that the generated plots are unique per benchmark"""

    input_vars = param.List(
        default=None,
        doc="A list of ParameterizedSweep variables to perform a parameter sweep over",
    )
    result_vars = param.List(
        default=None,
        doc="A list of ParameterizedSweep results collect and plot.",
    )

    const_vars = param.List(
        default=None,
        doc="Variables to keep constant but are different from the default value",
    )

    meta_vars = param.List(
        default=None,
        doc="Meta variables such as recording time and repeat id",
    )
    all_vars = param.List(
        default=None,
        doc="Stores a list of both the input_vars and meta_vars that are used to define a unique hash for the input",
    )
    iv_time = param.List(
        default=[],
        class_=TimeSnapshot | TimeEvent,
        doc="A parameter to represent the sampling the same inputs over time as a scalar type",
    )

    iv_time_event = param.List(
        default=[],
        class_=TimeEvent,
        doc="A parameter to represent the sampling the same inputs over time as a discrete type",
    )

    over_time: param.Boolean(
        False, doc="A parameter to control whether the function is sampled over time"
    )
    name: str = param.String(None, doc="The name of the benchmarkCfg")
    title: str = param.String(None, doc="The title of the benchmark")
    raise_duplicate_exception: str = param.Boolean(
        False, doc="Use this while debugging if filename generation is unique"
    )
    bench_name: str = param.String(
        None, doc="The name of the benchmark and the name of the save folder"
    )
    description: str = param.String(
        None,
        doc="A place to store a longer description of the function of the benchmark",
    )
    post_description: str = param.String(None, doc="A place to comment on the output of the graphs")

    has_results: bool = param.Boolean(
        False,
        doc="If this config has results, true, otherwise used to store titles and other bench metadata",
    )

    pass_repeat: bool = param.Boolean(
        False,
        doc="By default do not pass the kwarg 'repeat' to the benchmark function.  Set to true if you want the benchmark function to be passed the repeat number",
    )

    tag: str = param.String(
        "",
        doc="Use tags to group different benchmarks together. By default benchmarks are considered distinct from eachother and are identified by the hash of their name and inputs, constants and results and tag, but you can optionally change the hash value to only depend on the tag.  This way you can have multiple unrelated benchmarks share values with eachother based only on the tag value.",
    )

    hash_value: str = param.String(
        "",
        doc="store the hash value of the config to avoid having to hash multiple times",
    )

    def __init__(self, **params):
        super().__init__(**params)
        self.studies = []
        self.ds = xr.Dataset()
        self.plot_lib = None
        self.hmap = {}
        self.hmap_kdims = None
        self.iv_repeat = None

    def hash_persistent(self, include_repeats) -> str:
        """override the default hash function becuase the default hash function does not return the same value for the same inputs.  It references internal variables that are unique per instance of BenchCfg

        Args:
            include_repeats (bool) : by default include repeats as part of the hash execpt with using the sample cache
        """

        if include_repeats:
            # needed so that the historical xarray arrays are the same size
            repeats_hash = hash_sha1(self.repeats)
        else:
            repeats_hash = 0

        hash_val = hash_sha1(
            (
                hash_sha1(str(self.bench_name)),
                hash_sha1(str(self.title)),
                hash_sha1(self.over_time),
                repeats_hash,
                hash_sha1(self.debug),
                hash_sha1(self.tag),
            )
        )
        all_vars = self.input_vars + self.result_vars
        for v in all_vars:
            hash_val = hash_sha1((hash_val, v.hash_persistent()))

        for v in self.const_vars:
            hash_val = hash_sha1((v[0].hash_persistent(), hash_sha1(v[1])))

        return hash_val

    def get_optimal_value_indices(self, result_var: bch.ParametrizedSweep) -> xr.DataArray:
        """Get an xarray mask of the values with the best values found during a parameter sweep

        Args:
            result_var (bch.ParametrizedSweep): Optimal value of this result variable

        Returns:
            xr.DataArray: xarray mask of optimal values
        """
        result_da = self.ds[result_var.name]
        if result_var.direction == OptDir.maximize:
            opt_val = result_da.max()
        else:
            opt_val = result_da.min()
        indicies = result_da.where(result_da == opt_val, drop=True).squeeze()
        logging.info(f"optimal value of {result_var.name}: {opt_val.values}")
        return indicies

    def get_optimal_inputs(
        self, result_var: bch.ParametrizedSweep, keep_existing_consts: bool = True
    ) -> Tuple[bch.ParametrizedSweep, Any]:
        """Get a list of tuples of optimal variable names and value pairs, that can be fed in as constant values to subsequent parameter sweeps

        Args:
            result_var (bch.ParametrizedSweep): Optimal values of this result variable
            keep_existing_consts (bool): Include any const values that were defined as part of the parameter sweep

        Returns:
            Tuple[bch.ParametrizedSweep, Any]: Tuples of variable name and optimal values
        """
        da = self.get_optimal_value_indices(result_var)
        if keep_existing_consts:
            output = copy.deepcopy(self.const_vars)
        else:
            output = []

        for iv in self.input_vars:
            # assert da.coords[iv.name].values.size == (1,)
            if da.coords[iv.name].values.size == 1:
                # https://stackoverflow.com/questions/773030/why-are-0d-arrays-in-numpy-not-considered-scalar
                # use [()] to convert from a 0d numpy array to a scalar
                output.append((iv, da.coords[iv.name].values[()]))
            else:
                logging.warning(f"values size: {da.coords[iv.name].values.size}")
                output.append((iv, max(da.coords[iv.name].values[()])))

            logging.info(f"Maximum value of {iv.name}: {output[-1][1]}")
        return output

    def get_optimal_vec(
        self,
        result_var: bch.ParametrizedSweep,
        input_vars: List[bch.ParametrizedSweep],
    ) -> List[Any]:
        """Get the optimal values from the sweep as a vector.

        Args:
            result_var (bch.ParametrizedSweep): Optimal values of this result variable
            input_vars (List[bch.ParametrizedSweep]): Define which input vars values are returned in the vector

        Returns:
            List[Any]: A vector of optimal values for the desired input vector
        """

        da = self.get_optimal_value_indices(result_var)
        output = []
        for iv in input_vars:
            if da.coords[iv.name].values.size == 1:
                # https://stackoverflow.com/questions/773030/why-are-0d-arrays-in-numpy-not-considered-scalar
                # use [()] to convert from a 0d numpy array to a scalar
                output.append(da.coords[iv.name].values[()])
            else:
                logging.warning(f"values size: {da.coords[iv.name].values.size}")
                output.append(max(da.coords[iv.name].values[()]))
            logging.info(f"Maximum value of {iv.name}: {output[-1]}")
        return output

    def get_dataframe(self, reset_index=True) -> DataFrame:
        """Get the xarray results as a pandas dataframe

        Returns:
            pd.DataFrame: The xarray results array as a pandas dataframe
        """
        ds = self.ds.to_dataframe()
        if reset_index:
            return ds.reset_index()
        return ds

    def get_best_trial_params(self, canonical=False):
        out = self.studies[0].best_trials[0].params
        if canonical:
            return hmap_canonical_input(out)
        return out

    def get_best_holomap(self):
        return self.hmap[self.get_best_trial_params(True)]

    def get_pareto_front_params(self):
        return [p.params for p in self.studies[0].trials]

    def to_hv_dataset(self, reduce: ReduceType = ReduceType.AUTO) -> hv.Dataset:
        """Generate a holoviews dataset from the xarray dataset.

        Args:
            reduce (ReduceType, optional): Optionally perform reduce options on the dataset.  By default the returned dataset will calculate the mean and standard devation over the "repeat" dimension so that the dataset plays nicely with most of the holoviews plot types.  Reduce.Sqeeze is used if there is only 1 repeat and you want the "reduce" variable removed from the dataset. ReduceType.None returns an unaltered dataset. Defaults to ReduceType.AUTO.

        Returns:
            hv.Dataset: results in the form of a holoviews dataset
        """
        ds = convert_dataset_bool_dims_to_str(self.ds)

        if reduce == ReduceType.AUTO:
            reduce = ReduceType.REDUCE if self.repeats > 1 else ReduceType.SQUEEZE

        result_vars_str = [r.name for r in self.result_vars]
        hvds = hv.Dataset(ds, vdims=result_vars_str)
        if reduce == ReduceType.REDUCE:
            return hvds.reduce(["repeat"], np.mean, np.std)
        if reduce == ReduceType.SQUEEZE:
            return hv.Dataset(ds.squeeze("repeat", drop=True), vdims=result_vars_str)
        return hvds

    def to(self, hv_type: hv.Chart, reduce: ReduceType = ReduceType.AUTO, **kwargs) -> hv.Chart:
        return self.to_hv_dataset(reduce).to(hv_type, **kwargs)

    def to_curve(self, reduce: ReduceType = ReduceType.AUTO) -> hv.Curve:
        ds = self.to_hv_dataset(reduce)
        pt = ds.to(hv.Curve)
        if self.repeats > 1:
            pt *= ds.to(hv.Spread).opts(alpha=0.2)
        return pt

    def to_error_bar(self) -> hv.Bars:
        return self.to_hv_dataset(ReduceType.REDUCE).to(hv.ErrorBars)

    def to_points(self, reduce: ReduceType = ReduceType.AUTO) -> hv.Points:
        ds = self.to_hv_dataset(reduce)
        pt = ds.to(hv.Points)
        if reduce:
            pt *= ds.to(hv.ErrorBars)
        return pt

    def to_scatter(self):
        return self.to_hv_dataset(ReduceType.REDUCE).to(hv.Scatter)

    def to_scatter_jitter(self) -> hv.Scatter:
        ds = self.to_hv_dataset(ReduceType.NONE)
        pt = ds.to(hv.Scatter).opts(jitter=0.1).overlay("repeat").opts(show_legend=False)
        return pt

    def to_bar(self, reduce: ReduceType = ReduceType.AUTO) -> hv.Bars:
        ds = self.to_hv_dataset(reduce)
        pt = ds.to(hv.Bars)
        if reduce:
            pt *= ds.to(hv.ErrorBars)
        return pt

    def to_heatmap(self, reduce: ReduceType = ReduceType.AUTO, **kwargs) -> hv.HeatMap:
        return self.to(hv.HeatMap, reduce, **kwargs)

    def to_nd_layout(self) -> hv.NdLayout:
        return hv.NdLayout(self.hmap, kdims=self.hmap_kdims).opts(
            shared_axes=False, shared_datasource=False
        )

    def to_holomap(self) -> hv.HoloMap:
        # return hv.HoloMap(self.hmap, self.hmap_kdims)
        return hv.HoloMap(self.to_nd_layout()).opts(shared_axes=False)

    def to_dynamic_map(self) -> hv.DynamicMap:
        """use the values stored in the holomap dictionary to populate a dynamic map. Note that this is much faster than passing the holomap to a holomap object as the values are calculated on the fly"""

        def cb(**kwargs):
            return self.hmap[hmap_canonical_input(kwargs)].opts(framewise=True, shared_axes=False)

        kdims = []
        for i in self.input_vars + [self.iv_repeat]:
            kdims.append(i.as_dim(compute_values=True, debug=self.debug))

        return hv.DynamicMap(cb, kdims=kdims)

    def to_grid(self, inputs=None):
        if inputs is None:
            inputs = self.inputs_as_str()
        if len(inputs) > 2:
            inputs = inputs[:2]
        return self.to_holomap().grid(inputs)

    def to_table(self):
        return self.to(hv.Table, ReduceType.SQUEEZE)

    def inputs_as_str(self) -> List[str]:
        return [i.name for i in self.input_vars]

    def describe_sweep(self) -> pn.pane.Markdown:
        """Produce a markdown summary of the sweep settings

        Returns:
            pn.pane.Markdown: _description_
        """
        return pn.pane.Markdown(
            describe_benchmark(self, self.summarise_constant_inputs), label=self.bench_name
        )

    def summarise_sweep(self, name=None, describe=True) -> pn.pane.Markdown:
        """Produce panel output summarising the title, description and sweep setting"""
        if name is None:
            name = self.title
        col = pn.Column(name=name)
        col.append(pn.pane.Markdown(f"# {self.title}"))
        if self.description is not None:
            col.append(pn.pane.Markdown(self.description))
        if describe:
            col.append(self.describe_sweep())
        return col

    def to_optuna(self):
        from bencher.optuna_conversions import collect_optuna_plots

        return collect_optuna_plots(self)

    def optuna_targets(self) -> List[str]:
        target_names = []
        for rv in self.result_vars:
            if rv.direction != OptDir.none:
                target_names.append(rv.name)
        return target_names


def describe_benchmark(bench_cfg: BenchCfg, summarise_constant_inputs) -> str:
    """Generate a string summary of the inputs and results from a BenchCfg

    Args:
        bench_cfg (BenchCfg): BenchCfg to generate a summary of

    Returns:
        str: summary of BenchCfg
    """
    benchmark_sampling_str = ["```text"]
    benchmark_sampling_str.append("")
    benchmark_sampling_str.append("Input Variables:")
    for iv in bench_cfg.input_vars:
        benchmark_sampling_str.extend(describe_variable(iv, bench_cfg.debug, True))

    if bench_cfg.const_vars and (bench_cfg.summarise_constant_inputs or summarise_constant_inputs):
        benchmark_sampling_str.append("\nConstants:")
        for cv in bench_cfg.const_vars:
            benchmark_sampling_str.extend(describe_variable(cv[0], False, False, cv[1]))

    print_meta = True
    if len(bench_cfg.meta_vars) == 1:
        mv = bench_cfg.meta_vars[0]
        if mv.name == "repeat" and mv.samples == 1:
            print_meta = False

    if print_meta:
        benchmark_sampling_str.append("\nMeta Variables:")
        for mv in bench_cfg.meta_vars:
            benchmark_sampling_str.extend(describe_variable(mv, bench_cfg.debug, True))

    benchmark_sampling_str.append("\nResult Variables:")
    for rv in bench_cfg.result_vars:
        benchmark_sampling_str.extend(describe_variable(rv, bench_cfg.debug, False))

    benchmark_sampling_str.append("```")

    benchmark_sampling_str = "\n".join(benchmark_sampling_str)
    return benchmark_sampling_str


def convert_dataset_bool_dims_to_str(dataset: xr.Dataset) -> xr.Dataset:
    """Given a dataarray that contains boolean coordinates, conver them to strings so that holoviews loads the data properly

    Args:
        dataarray (xr.DataArray): dataarray with boolean coordinates

    Returns:
        xr.DataArray: dataarray with boolean coordinates converted to strings
    """
    bool_coords = {}
    for c in dataset.coords:
        if dataset.coords[c].dtype == bool:
            bool_coords[c] = [str(vals) for vals in dataset.coords[c].values]

    if len(bool_coords) > 0:
        return dataset.assign_coords(bool_coords)
    return dataset


class DimsCfg:
    """A class to store data about the sampling and result dimensions"""

    def __init__(self, bench_cfg: BenchCfg) -> None:
        self.dims_name = [i.name for i in bench_cfg.all_vars]

        self.dim_ranges = []
        self.dim_ranges = [i.values(bench_cfg.debug) for i in bench_cfg.all_vars]
        self.dims_size = [len(p) for p in self.dim_ranges]
        self.dim_ranges_index = [list(range(i)) for i in self.dims_size]
        self.dim_ranges_str = [f"{s}\n" for s in self.dim_ranges]
        self.coords = dict(zip(self.dims_name, self.dim_ranges))

        logging.debug(f"dims_name: {self.dims_name}")
        logging.debug(f"dim_ranges {self.dim_ranges_str}")
        logging.debug(f"dim_ranges_index {self.dim_ranges_index}")
        logging.debug(f"coords: {self.coords}")



"""This file has some examples for how to perform basic benchmarking parameter sweeps"""

from bencher.bencher import Bench, BenchRunCfg

# All the examples will be using the data structures and benchmark function defined in this file
from bencher.example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, bench_function


def example_1D_float(run_cfg: BenchRunCfg) -> Bench:
    """This example shows how to sample a 1 dimensional float variable and plot the result of passing that parameter sweep to the benchmarking function

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """

    bencher = Bench("benchmarking_example_float1D", bench_function, ExampleBenchCfgIn)

    # here we sample the input variable theta and plot the value of output1. The (noisy) function is sampled 20 times so you can see the distribution
    bencher.plot_sweep(
        title="Example 1D Float",
        input_vars=[ExampleBenchCfgIn.param.theta],
        result_vars=[ExampleBenchCfgOut.param.out_sin],
        description=example_1D_float.__doc__,
        run_cfg=run_cfg,
    )
    return bencher


if __name__ == "__main__":
    ex_run_cfg = BenchRunCfg()
    ex_run_cfg.repeats = 5
    ex_run_cfg.print_pandas = True
    ex_run_cfg.over_time = True

    example_1D_float(ex_run_cfg).show()



import logging
from copy import deepcopy

import panel as pn
import seaborn as sns

import bencher.plotting_functions as plt_func
from bencher.bench_cfg import BenchCfg, PltCfgBase, PltCntCfg
from bencher.variables.parametrised_sweep import ParametrizedSweep

from bencher.variables.results import ResultVar, ResultVec

import xarray as xr


class BenchPlotter:
    @staticmethod
    def plot(bench_cfg: BenchCfg, main_tab=None, append_cols=None) -> pn.pane:
        """Given the dataset result of a benchmark run, automatically dedeuce how to plot the data based on the types of variables that were sampled

        Args:
            bench_cfg (BenchCfg): Information on how the benchmark was sampled and the resulting data

        Returns:
            pn.pane: A panel containing plot results
        """

        if main_tab is None:
            main_tab = pn.Tabs(tabs_location="left")
        if bench_cfg.auto_plot:
            if len(bench_cfg.result_vars) == 0:
                tabs = pn.Column(name=bench_cfg.title)
                tabs.append(pn.pane.Markdown(f"{bench_cfg.description}"))

            else:
                plot_cols = bench_cfg.summarise_sweep(name="Plots View")
                if bench_cfg.over_time:
                    if len(bench_cfg.ds.coords["over_time"]) > 1:
                        plot_cols.append(pn.pane.Markdown("## Results Over Time"))
                        plot_cols.append(BenchPlotter.plot_results_row(bench_cfg))
                    else:
                        plot_cols.append(
                            pn.pane.Markdown(
                                "Results over time needs at least 2 time snapshots to plot"
                            )
                        )

                if bench_cfg.over_time:
                    plot_cols.append(pn.pane.Markdown("## Most Recent Results"))

                    bench_deep = deepcopy(bench_cfg)  # TODO do this in the future without copying
                    bench_deep.over_time = False
                    bench_deep.iv_time = []
                    last_time = bench_deep.ds.coords["over_time"][-1]
                    try:
                        bench_deep.ds = bench_deep.ds.sel(over_time=last_time)
                        plot_cols.append(BenchPlotter.plot_results_row(bench_deep))
                    except ValueError as e:
                        warning = f"failed to load historical data: {e}"
                        plot_cols.append(pn.pane.Markdown(warning))
                        logging.warning(warning)

                else:
                    plot_cols.append(pn.pane.Markdown("## Results"))

                    plot_cols.append(BenchPlotter.plot_results_row(bench_cfg))

                if bench_cfg.use_optuna:
                    plot_cols.extend(bench_cfg.to_optuna())

                if append_cols is not None:
                    plot_cols.extend(append_cols)
                # plot_cols.append(pn.Column(pn.Row()))#attempt to add spacer to stop overlapping but does not work todo

                plot_cols.append(pn.pane.Markdown(f"{bench_cfg.post_description}"))

                tabs = pn.Tabs(name=bench_cfg.title)
                tabs.append(plot_cols)

                if bench_cfg.serve_xarray:
                    tabs.append(
                        pn.Column(
                            pn.pane.Markdown(
                                """This page shows the with the inputs of the parameter sweep and the results in its native N-D xarray dataset format."""
                            ),
                            bench_cfg.ds,
                            name="Xarray Dataset View",
                        )
                    )
                if bench_cfg.serve_pandas:
                    tabs.append(
                        pn.Column(
                            pn.pane.Markdown(
                                """This page shows the with the inputs of the parameter sweep and the results as a pandas multiindex."""
                            ),
                            bench_cfg.ds.to_dataframe(),
                            name="Pandas Dataframe MultiIndex View",
                        )
                    )
                if bench_cfg.serve_pandas_flat:
                    tabs.append(
                        pn.Column(
                            pn.pane.Markdown(
                                """This page shows the with the inputs of the parameter sweep and the results as a flattened padas dataframe."""
                            ),
                            bench_cfg.get_dataframe(),
                            name="Pandas Dataframe Flattened View",
                        )
                    )
                # tabs.append(plot_cols)

            main_tab.append(tabs)

        main_tab.servable()
        return main_tab

    @staticmethod
    def plot_results_row(bench_cfg: BenchCfg) -> pn.Row:
        """Given a BenchCfg, plot each result variable and add to a panel row

        Args:
            bench_cfg (BenchCfg): The BenchCfg to plot

        Returns:
            pn.Row: A panel row with plots in it
        """
        # todo remove the scroll and make it resize dynamically
        plot_rows = pn.Row(name=bench_cfg.bench_name)

        plt_cnt_cfg = BenchPlotter.generate_plt_cnt_cfg(bench_cfg)

        for rv in bench_cfg.result_vars:
            plt_cnt_cfg.result_vars = 1
            if type(rv) == ResultVec:
                plt_cnt_cfg.vector_len = rv.size
            else:
                plt_cnt_cfg.vector_len = 1

            if bench_cfg.plot_lib is not None:
                print(f"float {plt_cnt_cfg.float_cnt}")
                print(f"cat {plt_cnt_cfg.cat_cnt}")
                print(f"vec {plt_cnt_cfg.vector_len}")
                plot_rows.append(bench_cfg.plot_lib.gather_plots(bench_cfg, rv, plt_cnt_cfg))
            # todo enable this check in future pr
            # if len(plot_rows) == 0:  # use the old plotting method as a backup
            plot_rows.append(
                pn.panel(BenchPlotter.plot_result_variable(bench_cfg, rv, plt_cnt_cfg))
            )

        return plot_rows

    @staticmethod
    def generate_plt_cnt_cfg(
        bench_cfg: BenchCfg,
    ) -> PltCntCfg:
        """Given a BenchCfg work out how many float and cat variables there are and store in a PltCntCfg class

        Args:
            bench_cfg (BenchCfg): See BenchCfg definition

        Raises:
            ValueError: If no plotting procedure could be automatically detected

        Returns:
            PltCntCfg: see PltCntCfg definition
        """
        plt_cnt_cfg = PltCntCfg()
        plt_cnt_cfg.float_vars = deepcopy(bench_cfg.iv_time)
        plt_cnt_cfg.cat_vars = []

        for iv in bench_cfg.input_vars:
            type_allocated = False
            typestr = str(type(iv))

            if "IntSweep" in typestr or "FloatSweep" in typestr:
                plt_cnt_cfg.float_vars.append(iv)
                type_allocated = True
            if "EnumSweep" in typestr or "BoolSweep" in typestr or "StringSweep" in typestr:
                plt_cnt_cfg.cat_vars.append(iv)
                type_allocated = True

            if not type_allocated:
                raise ValueError(f"No rule for type {typestr}")

        plt_cnt_cfg.float_cnt = len(plt_cnt_cfg.float_vars)
        plt_cnt_cfg.cat_cnt = len(plt_cnt_cfg.cat_vars)
        return plt_cnt_cfg

    @staticmethod
    def plot_result_variable(
        bench_cfg: BenchCfg, rv: ParametrizedSweep, plt_cnt_cfg: PltCntCfg
    ) -> pn.Column:
        """This method returns a single plot based on 1 result variable and a set of input variables.  It dedeuces the correct plot type by passing it to several configuration functions that operate on the number of inputs

        Args:
            bench_cfg (BenchCfg): A config of the input vars
            rv (ParametrizedSweep): a config of the result variable
            plt_cnt_cfg (PltCntCfg): A config of how many input types there are

        Raises:
            FileExistsError: If there are file naming conflicts

        Returns:
            PlotResult: A summary of all the data used to generate a plot
        """
        surf_col = pn.Column()

        sns_cfg = PltCfgBase()
        sns_cfg.y = rv.name  # by default the result variable is always plotted on the y axis

        if plt_cnt_cfg.float_cnt < 2:
            # set a marker for time series to its easy to see the measurment points
            if bench_cfg.over_time:
                sns_cfg.marker = "."
            if plt_cnt_cfg.float_cnt == 0:
                sns_cfg = BenchPlotter.plot_float_cnt_0(sns_cfg, plt_cnt_cfg)
            elif plt_cnt_cfg.float_cnt == 1:
                sns_cfg = BenchPlotter.plot_float_cnt_1(sns_cfg, plt_cnt_cfg)
            sns_cfg = BenchPlotter.get_axes_and_title(rv, sns_cfg, plt_cnt_cfg)
            surf_col.append(plt_func.plot_sns(bench_cfg, rv, sns_cfg))

        return surf_col

    @staticmethod
    def axis_mapping(cat_axis_order, sns_cfg: PltCfgBase, plt_cnt_cfg: PltCntCfg) -> PltCfgBase:
        """A function for determining the plot settings if there are 0 float variable and updates the PltCfgBase

        Args:
            sns_cfg (PltCfgBase): See PltCfgBase definition
            plt_cnt_cfg (PltCntCfg): See PltCntCfg definition

        Returns:
            PltCfgBase: See PltCfgBase definition
        """
        sns_dict = {}
        for i, v in enumerate(plt_cnt_cfg.cat_vars):
            axis = cat_axis_order[i]
            sns_dict[axis] = v.name

        sns_cfg.param.set_param(**sns_dict)

        return sns_cfg

    @staticmethod
    def get_axes_and_title(
        rv: ParametrizedSweep, sns_cfg: PltCfgBase, plt_cnt_cfg: PltCntCfg
    ) -> PltCntCfg:
        """Work out the axes label and plot tite

        Args:
            rv (ParametrizedSweep): result variable
            sns_cfg (PltCfgBase): plotting config
            plt_cnt_cfg (PltCntCfg): plot count config

        Returns:
            PltCfgBase: plot count config with titles added
        """
        all_vars = plt_cnt_cfg.float_vars + plt_cnt_cfg.cat_vars
        xvar = None
        for i in all_vars:
            if i.name == sns_cfg.x:
                xvar = i.units
        if xvar is not None:
            sns_cfg.xlabel = f"{sns_cfg.x} [{xvar}]"
        sns_cfg.ylabel = f"{sns_cfg.y} [{rv.units}]"
        sns_cfg.title = f"{sns_cfg.x} vs {sns_cfg.y}"
        return sns_cfg

    @staticmethod
    def plot_float_cnt_0(sns_cfg: PltCfgBase, plt_cnt_cfg: PltCntCfg) -> PltCfgBase:
        """A function for determining the plot settings if there are 0 float variable and updates the PltCfgBase

        Args:
            sns_cfg (PltCfgBase): See PltCfgBase definition
            plt_cnt_cfg (PltCntCfg): See PltCntCfg definition

        Returns:
            PltCfgBase: See PltCfgBase definition
        """

        if plt_cnt_cfg.float_cnt == 0:
            sns_cfg.plot_callback = sns.catplot
            sns_cfg.kind = "swarm"

            # as more cat variables are added, map them to these plot axes
            cat_axis_order = ["x", "row", "col", "hue"]
            sns_cfg = BenchPlotter.axis_mapping(cat_axis_order, sns_cfg, plt_cnt_cfg)

        return sns_cfg

    @staticmethod
    def plot_float_cnt_1(sns_cfg: PltCfgBase, plt_cnt_cfg: PltCntCfg) -> PltCfgBase:
        """A function for determining the plot settings if there is 1 float variable and updates the PltCfgBase

        Args:
            sns_cfg (PltCfgBase): See PltCfgBase definition
            plt_cnt_cfg (PltCntCfg): See PltCntCfg definition

        Returns:
            PltCfgBase: See PltCfgBase definition
        """

        if plt_cnt_cfg.float_cnt == 1:
            sns_cfg.x = plt_cnt_cfg.float_vars[0].name
            sns_cfg.kind = "line"
            sns_cfg.plot_callback = sns.relplot

            # as more cat variables are added, map them to these plot axes
            cat_axis_order = ["hue", "row", "col", "hue"]

            sns_cfg = BenchPlotter.axis_mapping(cat_axis_order, sns_cfg, plt_cnt_cfg)

        return sns_cfg

    @staticmethod
    def plot_float_cnt_2(plt_cnt_cfg: PltCntCfg, rv: ResultVar, debug: bool) -> PltCfgBase:
        """A function for determining the plot settings if there are 2 float variable and updates the PltCfgBase
        Args:
            sns_cfg (PltCfgBase): See PltCfgBase definition
            plt_cnt_cfg (PltCntCfg): See PltCntCfg definition
        Returns:
            PltCfgBase: See PltCfgBase definition
        """
        xr_cfg = PltCfgBase()
        if plt_cnt_cfg.float_cnt == 2:
            logging.info(f"surface plot: {rv.name}")
            xr_cfg.plot_callback_xra = xr.plot.plot
            xr_cfg.x = plt_cnt_cfg.float_vars[0].name
            xr_cfg.y = plt_cnt_cfg.float_vars[1].name
            xr_cfg.xlabel = f"{xr_cfg.x} [{plt_cnt_cfg.float_vars[0].units}]"
            xr_cfg.ylabel = f"{xr_cfg.y} [{plt_cnt_cfg.float_vars[1].units}]"
            xr_cfg.zlabel = f"{rv.name} [{rv.units}]"
            xr_cfg.title = f"{rv.name} vs ({xr_cfg.x} and {xr_cfg.y})"

            if plt_cnt_cfg.cat_cnt >= 1:
                logging.info("surface plot with 1 categorical")
                xr_cfg.row = plt_cnt_cfg.cat_vars[0].name
                xr_cfg.num_rows = len(plt_cnt_cfg.cat_vars[0].values(debug))
            if plt_cnt_cfg.cat_cnt >= 2:
                logging.info("surface plot with 2> categorical")
                xr_cfg.col = plt_cnt_cfg.cat_vars[1].name
                xr_cfg.num_cols = len(plt_cnt_cfg.cat_vars[1].values(debug))
        return xr_cfg



from typing import Optional, Tuple

import matplotlib.pyplot as plt
import pandas as pd
import panel as pn
import seaborn as sns

from bencher.plotting.plot_filter import PlotFilter, PlotInput, VarRange
from bencher.plotting.plot_types import PlotTypes
from bencher.plt_cfg import PltCfgBase


class Catplot:
    # shared plot filter for catplots
    float_0_cat_at_least1_vec_1_res_1 = PlotFilter(
        float_range=VarRange(0, 0),
        cat_range=VarRange(1, None),
        vector_len=VarRange(1, 1),
        result_vars=VarRange(1, 1),
    )

    @staticmethod
    def plot_setup(pl_in: PlotInput) -> Tuple[pd.DataFrame, PltCfgBase]:
        plt.figure(figsize=(4, 4))
        df = pl_in.bench_cfg.ds[pl_in.rv.name].to_dataframe().reset_index()
        sns_cfg = PltCfgBase()
        sns_cfg.x = pl_in.bench_cfg.input_vars[0].name
        x_units = pl_in.bench_cfg.input_vars[0].units
        sns_cfg.y = pl_in.rv.name
        sns_cfg.xlabel = f"{sns_cfg.x} [{x_units}]"
        sns_cfg.ylabel = f"{sns_cfg.y} [{pl_in.rv.units}]"
        sns_cfg.title = f"{sns_cfg.x} vs {sns_cfg.y}"
        return df, sns_cfg

    @staticmethod
    def plot_postprocess(fg: plt.figure, sns_cfg: PltCfgBase, name: str) -> Optional[pn.panel]:
        fg.fig.suptitle(sns_cfg.title)
        fg.set_xlabels(label=sns_cfg.xlabel, clear_inner=True)
        fg.set_ylabels(label=sns_cfg.ylabel, clear_inner=True)
        plt.tight_layout()
        return pn.panel(fg.fig, name=name)

    def catplot_common(self, pl_in: PlotInput, kind: str, name: str) -> Optional[pn.panel]:
        if self.float_0_cat_at_least1_vec_1_res_1.matches(pl_in.plt_cnt_cfg):
            df, sns_cfg = self.plot_setup(pl_in)
            sns_cfg.kind = kind
            fg = sns.catplot(df, **sns_cfg.as_sns_args())
            return self.plot_postprocess(fg, sns_cfg, name)
        return None

    def swarmplot(self, pl_in: PlotInput) -> Optional[pn.panel]:
        return self.catplot_common(pl_in, "swarm", PlotTypes.swarmplot)

    def violinplot(self, pl_in: PlotInput) -> Optional[pn.panel]:
        return self.catplot_common(pl_in, "violin", PlotTypes.violinplot)

    def boxplot(self, pl_in: PlotInput) -> Optional[pn.panel]:
        return self.catplot_common(pl_in, "box", PlotTypes.boxplot)

    def barplot(self, pl_in: PlotInput) -> Optional[pn.panel]:
        return self.catplot_common(pl_in, "bar", PlotTypes.barplot)

    def boxenplot(self, pl_in: PlotInput) -> Optional[pn.panel]:
        return self.catplot_common(pl_in, "boxen", PlotTypes.boxenplot)



import pytest
import unittest
from hypothesis import given, settings, strategies as st
import bencher as bch
from datetime import datetime

from strenum import StrEnum
from enum import auto
from typing import List
from param import Parameter
from itertools import combinations


class Enum1(StrEnum):
    """A generic enum"""

    enum1_val1 = auto()
    enum1_val2 = auto()


class Enum2(StrEnum):
    """Another generic enum"""

    enum2_val1 = auto()
    enum2_val2 = auto()


class BenchCfgTest(bch.ParametrizedSweep):
    """A class for representing all types of input"""

    float1 = bch.FloatSweep(default=0, bounds=[0, 1], doc="generic float 1", samples=2)
    float2 = bch.FloatSweep(default=0, bounds=[0, 1], doc="generic float 2", samples=2)
    int1 = bch.IntSweep(default=0, bounds=[0, 2], doc="generic int 1")
    int2 = bch.IntSweep(default=0, bounds=[0, 2], doc="generic int 2")
    bool1 = bch.BoolSweep(doc="generic bool 1")
    bool2 = bch.BoolSweep(doc="generic bool 2")
    enum1 = bch.EnumSweep(Enum1)
    enum2 = bch.EnumSweep(Enum2)


class BenchCfgTestOut(bch.ParametrizedSweep):
    """A class for representing all types of result"""

    out1 = bch.ResultVar(doc="generic result variable 1")
    out2 = bch.ResultVar(doc="generic result variable 2")
    outvec2 = bch.ResultVec(2, doc="A generic 2D vector")
    outvec3 = bch.ResultVec(3, doc="A generic 3D vector")


def bench_func(cfg: BenchCfgTest) -> BenchCfgTestOut:
    """A generic benchmark function"""
    output = BenchCfgTestOut()
    output.out1 = cfg.float1
    output.out2 = 2.0
    output.outvec2 = [0, 1]
    output.outvec3 = [0, 1, 2]
    return output


# all possible types of input
input_types = [
    BenchCfgTest.param.float1,
    BenchCfgTest.param.float2,
    BenchCfgTest.param.int1,
    BenchCfgTest.param.int2,
    BenchCfgTest.param.bool1,
    BenchCfgTest.param.bool2,
    BenchCfgTest.param.enum1,
    BenchCfgTest.param.enum2,
]

# all possible types of result
result_var_permutations = [
    [BenchCfgTestOut.param.out1],
    [BenchCfgTestOut.param.out1, BenchCfgTestOut.param.out2],
    # [BenchCfgTestOut.param.outvec2],
    # [BenchCfgTestOut.param.outvec3],
]


# the function used to generate all possible combination or permutations of input
generator_func = combinations

input_var_permutations = []
all_inputs = []

# all possible permutations of the input for a given number of inputs
for num_inputs in range(1, 3):
    input_var_permutations.extend([list(c) for c in generator_func(input_types, num_inputs)])


for p in input_var_permutations:
    print(",".join([pa.name for pa in p]))


@pytest.mark.skip
class TestAllCombinations(unittest.TestCase):
    """This class uses hypothesis to test as large a range as possible of input parameter combinations to make sure bencher always returns an error message rather than crashing.  After a long running parameter sweep the highest priority is to show as much data as possible even if some of the data processing or visulisations are not possible to calculate. (and result in an exception)"""

    def run_bencher_over_time(
        self,
        input_vars: List[Parameter],
        result_vars: List[bch.ResultVar],
        repeats: int,
    ):
        """Base function used to run benchers with a set of inputs,results and repeats over time"""
        bench = bch.Bench("test_bencher", bench_func, BenchCfgTest)

        for i in range(2):
            bench.plot_sweep(
                title="test_unique_filenames",
                input_vars=input_vars,
                result_vars=result_vars,
                run_cfg=bch.BenchRunCfg(
                    repeats=repeats,
                    over_time=True,
                    clear_history=i == 0,  # clear the history on the first iteration
                ),
                time_src=datetime(
                    1970, 1, i + 1
                ),  # repeatable time so outputs are same at the pixel level
            )

    @settings(deadline=10000, max_examples=50)
    @given(
        input_vars=st.sampled_from(input_var_permutations),
        result_vars=st.sampled_from(result_var_permutations),
        repeats=st.sampled_from([1, 2]),
    )
    def test_all_input_combinations_over_time_hyp(
        self,
        input_vars: List[Parameter],
        result_vars: List[bch.ResultVar],
        repeats: int,
    ):
        """Use hypothesis to enumerate combinations of inputs to bencher

        Args:
            input_vars (List[Parameter]): all possible sets of inputs
            result_vars (List[bch.ResultVar]): all possible sets of results
            repeats (int): 1 or 2 repeats (more than 2 repeats hits the same code as 2 repeats)
        """
        self.run_bencher_over_time(input_vars, result_vars, repeats)

    def test_falsifying_examples(self):
        """This test runs all the falsifying examples that were caught by hypothesis"""

        # TODO this has been been "fixed" by catching the pandas keyerrors for plot_surface_holo().  It needs to be fixed properly by investigating aggregation of bool datatypes.  At the moment bool varibles can cause agreggation errors. Possibly convert the bool to an enum type??
        self.run_bencher_over_time(
            [
                BenchCfgTest.param.float1,
                BenchCfgTest.param.enum1,
                BenchCfgTest.param.bool1,
            ],
            [BenchCfgTestOut.param.out1],
            1,
        )

        # Properly fixed
        self.run_bencher_over_time(
            [
                BenchCfgTest.param.bool1,
                BenchCfgTest.param.bool2,
                BenchCfgTest.param.enum1,
                BenchCfgTest.param.enum2,
            ],
            [BenchCfgTestOut.param.out1],
            1,
        )

        # TODO, These inputs need to be fixed.
        # self.run_bencher_over_time(
        #     [BenchCfgTest.param.float1],
        #     [BenchCfgTestOut.param.outvec3],
        #     1,
        # )

        # self.run_bencher_over_time(
        #     [BenchCfgTest.param.float1, BenchCfgTest.param.float2],
        #     [BenchCfgTestOut.param.outvec2],
        #     1,
        # )



"""This file has some examples for how to modify what plots are shown with PlotLibrary"""

import bencher as bch

# All the examples will be using the data structures and benchmark function defined in this file
from bencher.example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, bench_function


def example_plot_library(run_cfg: bch.BenchRunCfg) -> bch.Bench:
    """This example shows how use PlotLibrary.all() to display all possible plot types for a given benchmark

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """

    bencher = bch.Bench(
        "benchmarking_example_categorical1D", bench_function, ExampleBenchCfgIn, run_cfg=run_cfg
    )

    plot_lib = bch.PlotLibrary.all()

    bencher.plot_sweep(
        title="Example 1D Categorical",
        input_vars=[ExampleBenchCfgIn.param.postprocess_fn],
        const_vars=[ExampleBenchCfgIn.param.noisy.with_const(True)],
        result_vars=[ExampleBenchCfgOut.param.out_cos, ExampleBenchCfgOut.param.out_sin],
        description=example_plot_library.__doc__,
        run_cfg=run_cfg,
        plot_lib=plot_lib,
    )

    bencher.plot_sweep(
        title="Example 1D Float",
        input_vars=[ExampleBenchCfgIn.param.theta],
        const_vars=[ExampleBenchCfgIn.param.noisy.with_const(True)],
        result_vars=[ExampleBenchCfgOut.param.out_cos, ExampleBenchCfgOut.param.out_sin],
        description=example_plot_library.__doc__,
        run_cfg=run_cfg,
        plot_lib=plot_lib,
    )

    bencher.plot_sweep(
        title="Example Float Cat Single",
        input_vars=[ExampleBenchCfgIn.param.theta, ExampleBenchCfgIn.param.postprocess_fn],
        const_vars=[ExampleBenchCfgIn.param.noisy.with_const(True)],
        result_vars=[ExampleBenchCfgOut.param.out_cos, ExampleBenchCfgOut.param.out_sin],
        description=example_plot_library.__doc__,
        run_cfg=bch.BenchRunCfg(repeats=1),
        plot_lib=plot_lib,
    )

    # WIP
    # bencher.plot_sweep(
    #     title="Example Float Cat Repeats",
    #     input_vars=[ExampleBenchCfgIn.param.theta, ExampleBenchCfgIn.param.postprocess_fn],
    #     const_vars=[ExampleBenchCfgIn.param.noisy.with_const(True)],
    #     result_vars=[ExampleBenchCfgOut.param.out_cos, ExampleBenchCfgOut.param.out_sin],
    #     description=example_plot_library.__doc__,
    #     run_cfg=run_cfg,
    #     plot_lib=plot_lib,
    # )

    return bencher


if __name__ == "__main__":
    example_plot_library(bch.BenchRunCfg(repeats=5)).show()



import unittest

import panel as pn

import bencher as bch
from bencher.example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, bench_function


class TestPlotsCommon(unittest.TestCase):
    def create_bench_cfg(self, plot_name: str) -> bch.BenchCfg:
        """Given a plot name, run a sweep and plot with that plotting function

        Args:
            plot_name (str): name of the plot to use

        Returns:
            BenchCfg: Results with plots
        """
        bencher = bch.Bench("test_catplot", bench_function, ExampleBenchCfgIn)
        return bencher.plot_sweep(
            title="Test Cat Plots",
            input_vars=[ExampleBenchCfgIn.param.postprocess_fn],
            const_vars=[ExampleBenchCfgIn.param.noisy.with_const(True)],
            result_vars=[ExampleBenchCfgOut.param.out_cos],
            plot_lib=bch.PlotLibrary.none().add(plot_name),
            run_cfg=bch.BenchRunCfg(auto_plot=False),
        )

    def basic_plot_asserts(self, result: bch.BenchCfg, plot_name: str) -> None:
        """Check that the expected plots are generated

        Args:
            result (bch.BenchCfg): bench result to check
            plot_name (str): expected name of the plot
        """

        self.assertIsInstance(result, pn.viewable.Viewable)
        self.assertEqual(result.name, plot_name)



from bencher.plotting.plot_collection import PlotInput
from bencher.plotting.plot_types import PlotTypes
from bencher.plotting.plots.catplot import Catplot
from bencher.bench_cfg import PltCntCfg
from hypothesis import given, settings, strategies as st

from bencher.example.benchmark_data import ExampleBenchCfgOut


from .test_plots_common import TestPlotsCommon


class TestCatPlot(TestPlotsCommon):
    @settings(deadline=10000)
    @given(
        st.sampled_from(
            [PlotTypes.swarmplot, PlotTypes.barplot, PlotTypes.boxenplot, PlotTypes.violinplot]
        )
    )
    def test_plot_name(self, plot_name) -> None:
        bench_cfg = self.create_bench_cfg(plot_name)

        plt_cnt_cfg = PltCntCfg(float_cnt=0, cat_cnt=1)
        pl_in = PlotInput(bench_cfg, ExampleBenchCfgOut.param.out_cos, plt_cnt_cfg)
        cp = Catplot()
        plot_fn = getattr(cp, plot_name)
        result = plot_fn(pl_in)

        self.basic_plot_asserts(result, plot_name)



from enum import Enum
from typing import List

import numpy as np
from param import Integer, Number, Selector
from bencher.variables.sweep_base import SweepBase, shared_slots

# pylint: disable=super-init-not-called


class BoolSweep(SweepBase, Selector):
    """A class to reprsent a parameter sweep of bools"""

    __slots__ = shared_slots

    def __init__(self, units: str = "ul", samples: int = None, samples_debug: int = 2, **params):
        Selector.__init__(self, **params)
        self.units = units
        self.objects = [
            True,
            False,
        ]
        if samples is None:
            self.samples = 2
        self.samples_debug = samples_debug

    def values(self, debug=False) -> List[bool]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        return self.indices_to_samples(self.samples_debug if debug else self.samples, self.objects)


class StringSweep(SweepBase, Selector):
    """A class to reprsent a parameter sweep of strings"""

    __slots__ = shared_slots

    def __init__(
        self,
        string_list: List[str],
        units: str = "ul",
        samples: int = None,
        samples_debug: int = 2,
        **params,
    ):
        Selector.__init__(self, string_list, instantiate=True, **params)
        self.units = units
        if samples is None:
            self.samples = len(self.objects)
        else:
            self.samples = samples
        self.samples_debug = min(self.samples, samples_debug)

    def values(self, debug=False) -> List[str]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        return self.indices_to_samples(self.samples_debug if debug else self.samples, self.objects)


class EnumSweep(SweepBase, Selector):
    """A class to reprsent a parameter sweep of enums"""

    __slots__ = shared_slots

    def __init__(
        self, enum_type: Enum | List[Enum], units="ul", samples=None, samples_debug=2, **params
    ):
        # The enum can either be an Enum type or a list of enums
        list_of_enums = isinstance(enum_type, list)
        if list_of_enums:
            selector_list = enum_type  # already a list of enums
        else:
            # create a list of enums from the enum type definition
            selector_list = list(enum_type)
        Selector.__init__(self, selector_list, **params)
        if not list_of_enums:  # Grab the docs from the enum type def
            self.doc = enum_type.__doc__
        self.units = units
        if samples is None:
            self.samples = len(self.objects)
        else:
            self.samples = samples
        self.samples_debug = min(self.samples, samples_debug)

    def values(self, debug=False) -> List[Enum]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        return self.indices_to_samples(self.samples_debug if debug else self.samples, self.objects)


class IntSweep(SweepBase, Integer):
    """A class to reprsent a parameter sweep of ints"""

    __slots__ = shared_slots + ["sample_values"]

    def __init__(self, units="ul", samples=None, samples_debug=2, sample_values=None, **params):
        Integer.__init__(self, **params)
        self.units = units
        self.samples_debug = samples_debug

        if sample_values is None:
            if samples is None:
                if self.bounds is None:
                    raise RuntimeError("You must define bounds for integer types")
                self.samples = 1 + self.bounds[1] - self.bounds[0]
            else:
                self.samples = samples
            self.sample_values = None
        else:
            self.sample_values = sample_values
            self.samples = len(self.sample_values)
            if "default" not in params:
                self.default = sample_values[0]

    def values(self, debug=False) -> List[int]:
        """return all the values for a parameter sweep.  If debug is true return the  list"""
        sample_values = (
            self.sample_values
            if self.sample_values is not None
            else list(range(int(self.bounds[0]), int(self.bounds[1] + 1)))
        )

        return self.indices_to_samples(self.samples_debug if debug else self.samples, sample_values)

    ###THESE ARE COPIES OF INTEGER VALIDATION BUT ALSO ALLOW NUMPY INT TYPES
    def _validate_value(self, val, allow_None):
        if callable(val):
            return

        if allow_None and val is None:
            return

        if not isinstance(val, (int, np.integer)):
            raise ValueError(
                "Integer parameter %r must be an integer, " "not type %r." % (self.name, type(val))
            )

    ###THESE ARE COPIES OF INTEGER VALIDATION BUT ALSO ALLOW NUMPY INT TYPES
    def _validate_step(self, val, step):
        if step is not None and not isinstance(step, (int, np.integer)):
            raise ValueError(
                "Step can only be None or an " "integer value, not type %r" % type(step)
            )


class FloatSweep(SweepBase, Number):
    """A class to represent a parameter sweep of floats"""

    __slots__ = shared_slots + ["sample_values"]

    def __init__(
        self, units="ul", samples=10, samples_debug=2, sample_values=None, step=None, **params
    ):
        Number.__init__(self, step=step, **params)
        self.units = units
        self.samples_debug = samples_debug

        self.sample_values = sample_values

        if sample_values is None:
            self.samples = samples
        else:
            self.samples = len(self.sample_values)
            if "default" not in params:
                self.default = sample_values[0]

    def values(self, debug=False) -> List[float]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        samps = self.samples_debug if debug else self.samples
        if self.sample_values is None:
            if self.step is None:
                return np.linspace(self.bounds[0], self.bounds[1], samps)

            return np.arange(self.bounds[0], self.bounds[1], self.step)
        if debug:
            indices = [
                int(i)
                for i in np.linspace(0, len(self.sample_values) - 1, self.samples_debug, dtype=int)
            ]
            return [self.sample_values[i] for i in indices]
        return self.sample_values



# THIS IS NOT A WORKING EXAMPLE YET

# pylint: disable=duplicate-code,unused-argument


import bencher as bch
import math
import random
import numpy as np
import holoviews as hv

from strenum import StrEnum
from enum import auto


class Function(StrEnum):
    fn_cos = auto()
    fn_sin = auto()
    fn_log = auto()
    fn_arctan = auto()

    def call(self, arg) -> float:
        """Calls the function defined by the name of the enum

        Returns:
            float: The result of calling the function defined by the enum
        """
        return getattr(np, self.removeprefix("fn_"))(arg)


class PlotFunctions(bch.ParametrizedSweep):
    phase = bch.FloatSweep(
        default=0, bounds=[0, math.pi], doc="Input angle", units="rad", samples=5
    )

    freq = bch.FloatSweep(default=1, bounds=[0, math.pi], doc="Input angle", units="rad", samples=5)

    theta = bch.FloatSweep(
        default=0, bounds=[0, math.pi], doc="Input angle", units="rad", samples=10
    )

    compute_fn = bch.EnumSweep(Function)

    fn_output = bch.ResultVar(units="v", doc="sin of theta with some noise")

    out_sum = bch.ResultVar(units="v", doc="The sum")

    def calc(self, plot=True, **kwargs) -> dict:
        self.update_params_from_kwargs(**kwargs)
        noise = 0.1

        self.fn_output = self.compute_fn.call(self.phase + self.freq * self.theta) + random.uniform(
            0, noise
        )

        return self.get_results_values_as_dict(holomap=self.plot_holo(plot))

    def plot_holo(self, plot=True) -> hv.core.ViewableElement:
        """Plots a generic representation of the object that is not a basic hv datatype. In this case its an image of the values of the object, but it could be any representation of the object, e.g. a screenshot of the object state"""
        if plot:
            pt = hv.Text(0, 0, f"{self.phase}\n{self.freq}\n {self.theta}")
            pt *= hv.Ellipse(0, 0, 1)
            return pt
        return None

    def calc_vec(self, **kwargs) -> dict:
        theta = self.param.theta.values()
        kwargs.pop("theta", 0)
        dat = [self.calc(plot=False, theta=i, **kwargs)["fn_output"] for i in theta]
        # print(dat)
        self.out_sum = sum(dat)
        pt = hv.Curve((theta, dat), "theta", "voltage")
        # pt = hv.Text(0, 0, f"{self.compute_fn}\n{self.phase}\n{self.freq}")
        # pt *= hv.Ellipse(0, 0, 1)
        return self.get_results_values_as_dict(holomap=pt)


def example_holosweep_tap(run_cfg: bch.BenchRunCfg) -> bch.Bench:
    wv = PlotFunctions()

    run_cfg.use_optuna = True
    bench = bch.Bench("waves", wv.calc, plot_lib=None)

    res = bench.plot_sweep(
        "phase",
        input_vars=[wv.param.theta, wv.param.freq, wv.param.phase],
        result_vars=[wv.param.fn_output],
        run_cfg=run_cfg,
    )

    print(res.get_best_trial_params())
    bench.append(res.get_best_holomap())

    heatmap = res.to_heatmap().opts(tools=["hover", "tap"])
    posxy = hv.streams.Tap(source=heatmap, x=0, y=0)
    sld1 = wv.param.phase.as_slider(run_cfg.debug)

    def tap_plot(x, y):
        print(x, y)
        selres = bch.get_nearest_coords(res.ds, theta=x, freq=y, phase=sld1.value, repeat=1)
        return res.hmap[bch.hmap_canonical_input(selres)]

    tap_dmap = hv.DynamicMap(tap_plot, streams=[posxy])

    bench.append_tab(heatmap + tap_dmap, "Interactive Heatmap")

    bench.append(sld1)

    bench.append_tab(res.to_curve(), "Slider view")

    return bench


if __name__ == "__main__":
    example_holosweep_tap(bch.BenchRunCfg()).show()


# todo  https://discourse.holoviz.org/t/pointdraw-as-parameterized-class/3539



import panel as pn

from bencher.plotting.plot_filter import PlotInput
from bencher.plotting.plot_types import PlotTypes


class Tables:
    """A class to display the result data in tabular form"""

    def dataframe_flat(self, pl_in: PlotInput) -> pn.panel:
        """Returns a list of panel objects containing a flat dataframe."""
        df = pl_in.bench_cfg.get_dataframe()
        return pn.pane.DataFrame(df, name=PlotTypes.dataframe_flat)

    def dataframe_multi_index(self, pl_in: PlotInput) -> pn.panel:
        """Returns a list of panel objects containing a multi-index dataframe."""
        df = pl_in.bench_cfg.ds.to_dataframe()
        return pn.pane.DataFrame(df, name=PlotTypes.dataframe_multi_index)

    def dataframe_mean(self, pl_in: PlotInput) -> pn.panel:
        """Returns a list of panel objects containing a mean dataframe."""
        df = pl_in.bench_cfg.ds.mean("repeat").to_dataframe().reset_index()
        return pn.pane.DataFrame(df, name=PlotTypes.dataframe_mean)

    def xarray(self, pl_in: PlotInput) -> pn.panel:
        """Returns a list of panel objects containing an xarray object."""
        return pn.panel(pl_in.bench_cfg.ds, name=PlotTypes.xarray)



from setuptools import setup, find_packages

package_name = "bencher"

setup(
    name=package_name,
    version="0.0.1",
    description="A library for benchmarking code and generating reports for analysis",
    maintainer="austin.gregg-smith",
    maintainer_email="austin.gregg-smith@dyson.com",
    packages=find_packages(exclude=["test.*", "test"]),
    data_files=[
        ("share/ament_index/resource_index/packages", ["resource/" + package_name]),
        ("share/" + package_name, ["package.xml"]),
    ],
    install_requires=["setuptools"],
    zip_safe=False,
    test_suite="test",
    tests_require=["pytest"],
)



from .bencher import Bench, BenchCfg, BenchRunCfg
from .bench_runner import BenchRunner
from .plt_cfg import BenchPlotter, PltCfgBase
from .example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, bench_function
from .bench_plot_server import BenchPlotServer
from .variables.sweep_base import hash_sha1
from .variables.inputs import IntSweep, FloatSweep, StringSweep, EnumSweep, BoolSweep, SweepBase
from .variables.time import TimeSnapshot
from .variables.results import ResultVar, ResultVec, ResultHmap, OptDir
from .plotting.plot_library import PlotLibrary, PlotTypes
from .plotting.plot_filter import PlotInput, VarRange, PlotFilter
from .utils import hmap_canonical_input, get_nearest_coords, make_namedtuple
from .variables.parametrised_sweep import ParametrizedSweep
from .caching import CachedParams

from .optuna_conversions import to_optuna, summarise_study
from .bench_cfg import ReduceType



import unittest
from bencher.plotting.plot_filter import VarRange, PlotFilter
from hypothesis import given, strategies as st

from bencher.bench_cfg import PltCntCfg


class TestVarRange(unittest.TestCase):
    def test_matches_zero(self) -> None:
        zero_case = VarRange(0, 0)
        self.assertTrue(zero_case.matches(0))
        self.assertFalse(zero_case.matches(1))

    def test_matches_upto(self) -> None:
        var_range = VarRange(0, 1)
        # self.assertFalse(zero_case.matches(-1))
        self.assertTrue(var_range.matches(0))
        self.assertTrue(var_range.matches(1))
        self.assertFalse(var_range.matches(2))

    @given(st.integers())
    def test_default_matches_nothing(self, val) -> None:
        """Check that the default constructor produces a range that always returns false (so that any matches must be defined explicitly by the developer)"""
        var_range = VarRange()
        if val >= 0:
            self.assertFalse(var_range.matches(val))
        else:
            with self.assertRaises(ValueError):
                var_range.matches(val)

    @given(st.integers(min_value=0))
    def test_matches_none_upper(self, val) -> None:
        """Check that no upper bound matches all positive integers"""
        var_range = VarRange(0, None)
        self.assertTrue(var_range.matches(val))

    @given(st.integers(min_value=0))
    def test_matches_none_none(self, val) -> None:
        """No lower and no upper bounds should always match regardless of input"""
        var_range = VarRange(None, None)
        self.assertTrue(var_range.matches(val))


class TestPlotFilter(unittest.TestCase):
    @given(st.integers(min_value=0), st.integers(min_value=0))
    def test_matches_none(self, float_cnt, cat_cnt) -> None:
        """The default plot filter should not match any inputs"""

        null_case = PlotFilter()
        self.assertFalse(null_case.matches(PltCntCfg(float_cnt=float_cnt, cat_cnt=cat_cnt)))

    def test_matches_float(self) -> None:
        # match only float from 0 to 1
        pf = PlotFilter(float_range=VarRange(0, 1), cat_range=VarRange(None, None))

        self.assertTrue(pf.matches(PltCntCfg(float_cnt=0)))
        self.assertTrue(pf.matches(PltCntCfg(float_cnt=1)))
        self.assertFalse(pf.matches(PltCntCfg(float_cnt=2)))

    @given(st.integers(min_value=0))
    def test_matches_float_cat(self, cat_cnt) -> None:
        # match any cat count but only float from 0 to 1
        pf = PlotFilter(float_range=VarRange(0, 1), cat_range=VarRange(None, None))

        self.assertTrue(pf.matches(PltCntCfg(float_cnt=0, cat_cnt=cat_cnt)))
        self.assertTrue(pf.matches(PltCntCfg(float_cnt=1, cat_cnt=cat_cnt)))
        self.assertFalse(pf.matches(PltCntCfg(float_cnt=2, cat_cnt=cat_cnt)))



import holoviews as hv
import bencher as bch


class Investment(bch.ParametrizedSweep):
    inflation = bch.FloatSweep(default=8, bounds=[0.5, 15], units="%", samples=5)
    monthly_contribution = bch.FloatSweep(default=0, bounds=[0, 1000], units="$", samples=5)
    interest = bch.FloatSweep(default=6.0, bounds=(0, 10), units="%", samples=5)
    tax_rate = bch.IntSweep(default=0, bounds=(0, 40), samples=2)
    period = bch.IntSweep(default=12, bounds=[1, 12], units="month")
    starting_value = bch.FloatSweep(default=1000, bounds=(0, 2000), samples=3)

    portfolio_value = bch.ResultVar(default=0, units="$")

    def call(self, **kwargs) -> dict:
        self.update_params_from_kwargs(**kwargs)
        self.portfolio_value = self.starting_value
        values = []
        for i in range(self.period):
            self.portfolio_value += self.monthly_contribution
            new_interest = self.portfolio_value * (self.interest / 100.0)
            self.portfolio_value += new_interest * (1 + self.tax_rate)
            self.portfolio_value *= 1 - (self.inflation / 100.0)
            values.append((i, self.portfolio_value))
        return self.get_results_values_as_dict(
            hv.Curve(values, self.param.period.as_dim(), self.param.portfolio_value.as_dim())
        )


if __name__ == "__main__":
    inv = Investment()

    bench = bch.Bench("Investment", inv.call, plot_lib=None)

    res = bench.plot_sweep(
        "Investment Sweep",
        description="This calculator lets you calculate the value of your investment over time",
        input_vars=[inv.param.monthly_contribution, inv.param.interest, inv.param.inflation],
        result_vars=[inv.param.portfolio_value],
    )

    bench.append(res.to_grid())
    bench.append(res.to(hv.Table).opts(width=1000))

    bench.append_tab(Investment().to_dynamic_map(name="Investment"))

    bench.show()
